# -*- coding: utf8 -*-
from collections import defaultdict
from mock import MagicMock as Mock, patch
from pretend import stub

from helga.plugins import help


@patch('helga.plugins.help.registry')
def test_help_should_whisper(plugins):
    client = Mock()
    plugins.enabled_plugins = defaultdict(list)
    plugins.plugins = {}

    help.help(client, '#bots', 'me', 'foo', 'bar', [])

    assert client.me.called
    client.me.assert_called_with('#bots', 'whispers to me')


@patch('helga.plugins.help.registry')
def test_help_gets_object_help_attrs(plugins):
    snowman = u'☃'
    foo_plugin = stub(help='foo plugin')
    bar_plugin = stub()
    uni_plugin = stub(help=u'unicode plugin {0}'.format(snowman))

    client = Mock()
    plugins.enabled_plugins = {
        '#bots': ['foo', 'bar', snowman],
    }
    plugins.plugins = {
        'foo': foo_plugin,
        'bar': bar_plugin,
        snowman: uni_plugin,
    }

    help.help(client, '#bots', 'me', 'help', 'help', [])

    nick, message = client.msg.call_args[0]
    assert nick == 'me'
    assert 'me, here are the plugins I know about' in message
    assert u'[{0}] unicode plugin {0}'.format(snowman) in message
    assert '[foo] foo plugin' in message


@patch('helga.plugins.help.registry')
def test_help_gets_decorated_help_attrs(plugins):
    snowman = u'☃'
    foo_plugin = stub(help='foo plugin')
    foo_fn = stub(_plugins=[foo_plugin])

    uni_plugin = stub(help=u'unicode plugin {0}'.format(snowman))
    uni_fn = stub(_plugins=[uni_plugin])

    client = Mock()
    plugins.enabled_plugins = {
        '#bots': ['foo', snowman],
    }
    plugins.plugins = {
        'foo': foo_fn,
        snowman: uni_fn,
    }

    help.help(client, '#bots', 'me', 'help', 'help', [])

    nick, message = client.msg.call_args[0]
    assert nick == 'me'
    assert 'me, here are the plugins I know about' in message
    assert u'[{0}] unicode plugin {0}'.format(snowman) in message
    assert '[foo] foo plugin' in message


@patch('helga.plugins.help.registry')
def test_help_gets_multi_decorated_help_attrs(plugins):
    foo_plugin = stub(help='foo plugin')
    bar_plugin = stub(help='bar plugin')
    uni_plugin = stub(help=u'unicode plugin ☃')
    foo_fn = stub(_plugins=[foo_plugin, bar_plugin, uni_plugin])

    client = Mock()
    plugins.enabled_plugins = {'#bots': ['foo']}
    plugins.plugins = {'foo': foo_fn}

    help.help(client, '#bots', 'me', 'help', 'help', [])

    nick, message = client.msg.call_args[0]
    assert nick == 'me'
    assert 'me, here are the plugins I know about' in message
    assert u'[foo] foo plugin. bar plugin. unicode plugin ☃' in message


@patch('helga.plugins.help.registry')
def test_help_gets_single_plugin(plugins):
    snowman = u'☃'
    foo_plugin = stub(help='foo plugin')
    bar_plugin = stub(help='bar plugin')
    uni_plugin = stub(help=u'unicode plugin {0}'.format(snowman))

    client = Mock()
    plugins.enabled_plugins = {
        '#bots': ['foo', 'bar', snowman],
    }
    plugins.plugins = {
        'foo': foo_plugin,
        'bar': bar_plugin,
        snowman: uni_plugin,
    }

    message = help.help(client, '#bots', 'me', 'help', 'help', ['bar'])
    assert '[bar] bar plugin' in message
    assert '[foo] foo plugin' not in message
    assert u'[{0}] unicode plugin {0}'.format(snowman) not in message

    message = help.help(client, '#bots', 'me', 'help', 'help', [snowman])
    assert '[bar] bar plugin' not in message
    assert '[foo] foo plugin' not in message
    assert u'[{0}] unicode plugin {0}'.format(snowman) in message


@patch('helga.plugins.help.registry')
def test_help_single_returns_unknown(plugins):
    foo_plugin = stub(help='foo plugin')
    bar_plugin = stub(help='bar plugin')

    client = Mock()
    plugins.enabled_plugins = {
        '#bots': ['foo', 'bar'],
    }
    plugins.plugins = {
        'foo': foo_plugin,
        'bar': bar_plugin,
    }

    resp = help.help(client, '#bots', 'me', 'help', 'help', ['baz'])
    assert resp == "Sorry me, I don't know about that plugin"
    assert not client.msg.called

    # Can handle unicode
    resp = help.help(client, '#bots', 'me', 'help', 'help', [u'☃'])
    assert resp == "Sorry me, I don't know about that plugin"
    assert not client.msg.called


@patch('helga.plugins.help.registry')
def test_help_with_unloaded_plugin(plugins):
    snowman = u'☃'
    foo_plugin = stub(help='foo plugin')

    client = Mock()
    plugins.enabled_plugins = {'#bots': ['foo', snowman]}
    plugins.plugins = {'foo': foo_plugin}

    resp = help.help(client, '#bots', 'me', 'help', 'help', [snowman])
    assert resp == u"Sorry me, there's no help string for plugin '{0}'".format(snowman)
    assert not client.msg.called


@patch('helga.plugins.help.registry')
def test_help_object_with_empty_help_attr(plugins):
    foo_plugin = stub(help='')
    uni_plugin = stub(help='')

    client = Mock()
    plugins.enabled_plugins = {'#bots': ['foo', u'☃']}
    plugins.plugins = {'foo': foo_plugin, u'☃': uni_plugin}

    help.help(client, '#bots', 'me', 'help', 'help', [])

    _, message = client.msg.call_args[0]
    assert '[foo] No help string for this plugin' in message
    assert u'[☃] No help string for this plugin' in message


@patch('helga.plugins.help.registry')
def test_help_decorated_with_empty_help_attr(plugins):
    foo_fn = stub(_plugins=[stub(help='')])
    uni_fn = stub(_plugins=[stub(help='')])

    client = Mock()
    plugins.enabled_plugins = {'#bots': ['foo', u'☃']}
    plugins.plugins = {'foo': foo_fn, u'☃': uni_fn}

    help.help(client, '#bots', 'me', 'help', 'help', [])

    _, message = client.msg.call_args[0]
    assert '[foo] No help string for this plugin' in message
    assert u'[☃] No help string for this plugin' in message


def test_format_help_string():
    unistr = u'[☃] snowman. ☃'
    string = '[foo] just foo. bar'

    assert unistr == help.format_help_string(u'☃', 'snowman', u'☃')
    assert string == help.format_help_string('foo', 'just foo', 'bar')



import re

from helga.plugins import match


@match(r'^s/(.*?)/(.*?)(/.*?)?$')
def meant_to_say(client, channel, nick, message, matches):
    """
    A plugin so users can correct what they have said. For example::

        <sduncan> this is a foo message
        <sduncan> s/foo/bar
        <helga> sduncan meant to say: this is a bar message
    """
    try:
        last = client.last_message[channel][nick]
    except KeyError:
        return None

    old, new, reflags = matches[0]
    count = 1

    if re.search('g', reflags, re.I):
        count = 0
    if re.search('i', reflags, re.I):
        regex = re.compile(old, re.I)
    else:
        regex = re.compile(old)

    modified = regex.sub(new, last, count)

    # Don't respond if we don't replace anything ... it's annoying
    if modified != last:
        return '{0} meant to say: {1}'.format(nick, modified)



import warnings


from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from helga import settings


def _connect():
    db_settings = getattr(settings, 'DATABASE', {})

    try:
        client = MongoClient(db_settings['HOST'], db_settings['PORT'])
    except ConnectionFailure:
        warnings.warn('MongoDB is not available. Some features may not work')
        return None, None
    else:
        db = client[db_settings['DB']]

        if 'USERNAME' in db_settings and 'PASSWORD' in db_settings:
            db.authenticate(db_settings['USERNAME'], db_settings['PASSWORD'])

        return client, db


client, db = _connect()



# -*- coding: utf8 -*-
from datetime import datetime

from mock import patch
from pretend import stub

from helga.plugins import facts


def test_term_regex():
    pat = facts.term_regex('foo')

    assert bool(pat.match('foo'))
    assert not bool(pat.match('will not match foo'))


def test_term_regex_handles_unicode():
    snowman = u'☃'
    pat = facts.term_regex(snowman)
    assert bool(pat.match(snowman))
    assert not bool(pat.match('no snowman for you :('))


@patch('helga.plugins.facts.db')
def test_show_fact_no_record(db):
    db.facts.find_one.return_value = None
    assert facts.show_fact('foo') is None


@patch('helga.plugins.facts.db')
def test_show_fact_without_author(db):
    fact = {
        'term': 'foo',
        'fact': 'foo is bar',
    }

    db.facts.find_one.return_value = fact
    assert facts.show_fact('foo') == 'foo is bar'


@patch('helga.plugins.facts.db')
def test_show_fact_without_set_time(db):
    fact = {
        'term': 'foo',
        'fact': 'foo is bar',
        'set_by': 'sduncan',
    }

    db.facts.find_one.return_value = fact
    assert facts.show_fact('foo') == 'foo is bar (sduncan)'


@patch('helga.plugins.facts.db')
def test_show_fact(db):
    fact = {
        'term': 'foo',
        'fact': 'foo is bar',
        'set_by': 'sduncan',
        'set_date': 1360849874.686594,  # 02/14/2013 08:51AM
    }

    db.facts.find_one.return_value = fact
    assert facts.show_fact('foo') == 'foo is bar (sduncan on 02/14/2013 08:51AM)'


@patch('helga.plugins.facts.db')
def test_show_fact_handles_unicode(db):
    fact = {
        'term': u'☃',
        'fact': u'☃ is unicode snowman',
        'set_by': 'sduncan',
        'set_date': 1360849874.686594,  # 02/14/2013 08:51AM
    }

    db.facts.find_one.return_value = fact
    assert facts.show_fact(u'☃') == u'☃ is unicode snowman (sduncan on 02/14/2013 08:51AM)'


@patch('helga.plugins.facts.db')
def test_show_fact_set_date_is_datetime(db):
    fact = {
        'term': 'foo',
        'fact': 'foo is bar',
        'set_by': 'sduncan',
        'set_date': datetime(2013, 2, 14, 8, 51),
    }

    db.facts.find_one.return_value = fact
    assert facts.show_fact('foo') == 'foo is bar (sduncan on 02/14/2013 08:51AM)'


@patch('helga.plugins.facts.settings')
@patch('helga.plugins.facts.add_fact')
def test_facts_match_with_nick(add_fact, settings):
    settings.FACTS_REQUIRE_NICKNAME = True
    client = stub(nickname='helga')

    # The format the regex will use
    found = [('helga: foo bar', 'is', '', 'this is the response')]
    add_fact.return_value = 'ok'

    assert 'ok' == facts.facts_match(client, '', '', '', found)
    add_fact.assert_called_with('foo bar', 'foo bar is this is the response', '')


@patch('helga.plugins.facts.settings')
@patch('helga.plugins.facts.add_fact')
def test_facts_match_requires_nick(add_fact, settings):
    settings.FACTS_REQUIRE_NICKNAME = True
    client = stub(nickname='helga')

    # The format the regex will use
    found = [('foo bar', 'is', '', 'this is the response')]
    add_fact.return_value = 'ok'

    assert facts.facts_match(client, '', '', '', found) is None


def test_facts_match_found_is_show_fact():
    facts.show_fact = lambda s: s
    assert facts.facts_match(None, '', '', '', ['foo']) == 'foo'
    assert facts.facts_match(None, '', '', '', [u'☃']) == u'☃'


@patch('helga.plugins.facts.settings')
@patch('helga.plugins.facts.add_fact')
def test_facts_match_handles_unicode(add_fact, settings):
    add_fact.return_value = 'ok'
    settings.FACTS_REQUIRE_NICKNAME = False
    client = stub(nickname='helga')

    snowman = u'☃'
    found_args = [
        [(snowman, 'is', '', 'here is a snowman')],
        [(snowman, 'is', '<reply>', 'here is a snowman')],
    ]
    for args in found_args:
        assert 'ok' == facts.facts_match(client, '', '', '', args)

    # Handle unicode with nick checking
    settings.FACTS_REQUIRE_NICKNAME = True
    for args in found_args:
        # prepend the nick
        with_nick = [['helga ' + args[0][0]] + list(args[0][1:])]
        assert facts.facts_match(client, '', '', '', args) is None
        assert 'ok' == facts.facts_match(client, '', '', '', with_nick)


@patch('helga.plugins.facts.add_fact')
def test_facts_as_reply(add_fact):
    # The format the regex will use
    found = [('foo bar', 'is', '<reply>', 'this is the response')]
    add_fact.return_value = 'ok'

    assert 'ok' == facts.facts_match('', '', '', '', found)
    add_fact.assert_called_with('foo bar', 'this is the response', '')


@patch('helga.plugins.facts.add_fact')
def test_facts(add_fact):
    # The format the regex will use
    found = [('foo bar', 'is', '', 'this is the response')]
    add_fact.return_value = 'ok'

    assert 'ok' == facts.facts_match('', '', '', '', found)
    add_fact.assert_called_with('foo bar', 'foo bar is this is the response', '')


@patch('helga.plugins.facts.db')
def test_add_fact_does_nothing_when_found(db):
    db.facts.find.return_value = db
    db.count.return_value = 1
    facts.add_fact('foo', 'bar', 'baz')
    assert not db.facts.insert.called


@patch('helga.plugins.facts.db')
@patch('helga.plugins.facts.time')
def test_add_fact(time, db):
    time.time.return_value = 1
    db.facts.find.return_value = db
    db.count.return_value = 0
    facts.add_fact('foo', 'bar', 'baz')
    db.facts.insert.assert_called_with({
        'term': 'foo',
        'fact': 'bar',
        'set_by': 'baz',
        'set_date': 1,
    })


@patch('helga.plugins.facts.db')
def test_forget_fact(db):
    facts.forget_fact('foo')
    db.facts.remove.assert_called_with({
        'term': facts.term_regex('foo'),
    })


def test_facts_command():
    facts.forget_fact = lambda s: s
    args = ['foo', 'bar', 'baz']
    assert 'foo bar baz' == facts.facts_command(None, '#bots', 'me', '!forget foo', 'forget', args)
    assert facts.facts_command(None, '#bots', 'me', '!forget foo', 'foobar', args) is None


@patch('helga.plugins.facts.facts_match')
@patch('helga.plugins.facts.facts_command')
def test_facts_plugin(cmd, match):
    cmd.return_value = 'command'
    match.return_value = 'match'

    assert 'command' == facts.facts(None, '#bots', 'me', 'command', 'foo', 'bar')
    assert 'match' == facts.facts(None, '#bots', 'me', 'match', 'match')



import tweepy

from helga import log, settings


logger = log.getLogger(__name__)


def is_properly_configured():
    return all([
        getattr(settings, 'TWITTER_CONSUMER_KEY', None),
        getattr(settings, 'TWITTER_CONSUMER_SECRET', None),
        getattr(settings, 'TWITTER_OAUTH_TOKEN', None),
        getattr(settings, 'TWITTER_OAUTH_TOKEN_SECRET', None),
    ])


def message_max(message, max):
    if len(message) > max:
        logger.warning('Message exceeds %s characters. Truncating', max)
        message = message[:max]
    return message


def message_140(message):
    return message_max(message, 140)


def get_api():
    auth = tweepy.OAuthHandler(settings.TWITTER_CONSUMER_KEY, settings.TWITTER_CONSUMER_SECRET)
    auth.set_access_token(settings.TWITTER_OAUTH_TOKEN, settings.TWITTER_OAUTH_TOKEN_SECRET)
    return tweepy.API(auth)


def tweet(message):
    if not is_properly_configured():
        logger.error('Twitter API requires consumer key, consumer secret, oauth token, oauth secret')
        return

    message = message_140(message)

    try:
        logger.info('Tweeting: %s', message)
        status = get_api().update_status(message)
    except:  # pragma: no cover
        logger.exception('Could not post status')
    else:
        tweet_url = 'http://twitter.com/{0}/status/{1}'.format(settings.TWITTER_USERNAME, status.id)
        logger.info('Tweeted: %s', tweet_url)
        return tweet_url



import random
import re

from twisted.internet import reactor

from helga.plugins import command, preprocessor


silence_acks = (
    u'silence is golden',
    u'shutting up',
    u'biting my tongue',
    u'fine, whatever',
)

unsilence_acks = (
    u'speaking once again',
    u'did you miss me?',
    u'FINALLY',
    u'thanks {nick}, i was getting bored'
)

snarks = (
    u'why would you want to do that {nick}?',
    u'do you really despise me that much {nick}?',
    u'whatever i do what i want',
    u'no can do, i love the sound of my own voice',
)

# Set of silenced channels
silenced = set()


def auto_unsilence(client, channel, length):
    global silenced

    if channel in silenced:
        silenced.discard(channel)
        client.msg(channel, u"Speaking again after waiting {0} minutes".format(length//60))


@preprocessor
@command('stfu', aliases=['speak'],
         help="Tell the bot to be quiet or not. Usage: helga (speak|stfu [for <time_in_minutes>])")
def stfu(client, channel, nick, message, *args):
    global silenced

    # Handle the message preprocesor
    if len(args) == 0:
        # Duh, don't silence the speak command
        is_speak = bool(re.findall(ur'^{0}\W*\s(speak)$'.format(client.nickname), message))

        if channel in silenced and not is_speak:
            message = u''

        return channel, nick, message

    elif len(args) == 2:
        resp = u''

        if channel == nick:  # Private message
            resp = random.choice(snarks)
        elif args[0] == 'stfu':
            silenced.add(channel)
            cmdargs = args[1]
            resp = random.choice(silence_acks)

            if len(cmdargs) > 0 and cmdargs[0] == 'for':
                try:
                    length = int(cmdargs[1]) * 60
                except (TypeError, ValueError, IndexError):
                    pass
                else:
                    reactor.callLater(length, auto_unsilence, client, channel, length)
                    resp = u"OK {0}, I'll be back in {1} min".format(nick, cmdargs[1])
        elif args[0] == 'speak':
            if channel not in silenced:
                return None

            resp = random.choice(unsilence_acks)
            silenced.discard(channel)

        return resp.format(nick=nick)



# -*- coding: utf8 -*-
from helga.util.encodings import (from_unicode,
                                  from_unicode_args,
                                  to_unicode,
                                  to_unicode_args)


def test_to_unicode_with_unicode_string():
    snowman = u'☃'
    retval = to_unicode(snowman)
    assert snowman == retval
    assert isinstance(retval, unicode)


def test_to_unicode_with_byte_string():
    snowman = u'☃'
    bytes = '\xe2\x98\x83'
    retval = to_unicode(bytes)
    assert snowman == retval
    assert isinstance(retval, unicode)


def test_from_unicode_with_unicode_string():
    snowman = u'☃'
    bytes = '\xe2\x98\x83'
    retval = from_unicode(snowman)
    assert bytes == retval
    assert isinstance(retval, str)


def test_from_unicode_with_byte_string():
    bytes = '\xe2\x98\x83'
    retval = from_unicode(bytes)
    assert bytes == retval
    assert isinstance(retval, str)


def test_to_unicode_args():
    @to_unicode_args
    def foo(arg1, arg2, arg3):
        return arg1, arg2, arg3

    snowman = u'☃'
    bytes = '\xe2\x98\x83'
    # None here to ensure bad things don't happen
    retval = foo(bytes, 'foo', None)

    assert retval[0] == snowman
    assert retval[1] == u'foo'
    assert retval[2] is None
    assert isinstance(retval[0], unicode)
    assert isinstance(retval[1], unicode)


def test_from_unicode_args():
    @from_unicode_args
    def foo(arg1, arg2, arg3):
        return arg1, arg2, arg3

    snowman = u'☃'
    bytes = '\xe2\x98\x83'
    # None here to ensure bad things don't happen
    retval = foo(snowman, 'foo', None)

    assert retval[0] == bytes
    assert retval[1] == 'foo'
    assert retval[2] is None
    assert isinstance(retval[0], str)
    assert isinstance(retval[1], str)



import sys

import smokesignal

from twisted.internet import reactor, ssl

from helga import comm, settings


_help = """
Usage: helga <option>

Avalable options:
help        Show this message
start       Start the helga bot (CTRL-C to stop)
"""


def start():
    smokesignal.emit('started')

    factory = comm.Factory()
    if settings.SERVER.get('SSL', False):
        reactor.connectSSL(settings.SERVER['HOST'],
                           settings.SERVER['PORT'],
                           factory,
                           ssl.ClientContextFactory())
    else:
        reactor.connectTCP(settings.SERVER['HOST'],
                           settings.SERVER['PORT'],
                           factory)
    reactor.run()


def run():
    args = sys.argv[1:]
    if len(args) == 0 or args[0] == 'help':
        start()



# -*- coding: utf8 -*-
import re

import pytest

from mock import Mock, patch, call

from helga.plugins import poems, ResponseNotReady


@patch('helga.plugins.poems.get_random_line')
def test_fix_repitition_replaces(get_random_line):
    poem = ['foo', 'bar', 'foo']
    get_random_line.return_value = 'baz'
    poem = poems.fix_repitition(poem)

    assert poem == ['foo', 'bar', 'baz']


@patch('helga.plugins.poems.get_random_line')
def test_fix_repititions_gives_up_after_retry(get_random_line):
    poem = ['foo', 'bar', 'foo']
    get_random_line.return_value = 'foo'
    poem = poems.fix_repitition(poem)

    assert poem == ['foo', 'bar', 'foo']
    assert poems.get_random_line.call_count == 2


@patch('helga.plugins.poems.get_random_line')
def test_fix_repitition_does_not_replace(get_random_line):
    poem = ['foo', 'bar', 'baz']
    poem = poems.fix_repitition(poem)

    assert poem == ['foo', 'bar', 'baz']
    assert not poems.get_random_line.called


@patch('helga.plugins.poems.db')
def test_add(db):
    poems.add(5, 'foobar')
    assert db.haiku.insert.called


@patch('helga.plugins.poems.use')
@patch('helga.plugins.poems.add')
def test_add_use(add, use):
    poems.add_use(5, 'foo')
    add.assert_called_with(5, 'foo', author=None)
    use.assert_called_with(5, 'foo')


@patch('helga.plugins.poems.db')
def test_remove(db):
    poems.remove(5, 'foobar')
    assert db.haiku.remove.called


@patch('helga.plugins.poems.db')
def test_get_random_line(db):
    result = Mock()
    result.sort = result
    result.count.return_value = 1
    result.limit.return_value = result
    result.skip.return_value = result
    result.next.return_value = {'message': 'fives1'}

    fake_find = Mock(return_value=result)

    db.haiku.find = fake_find
    line = poems.get_random_line(5)
    assert line == 'fives1'


@patch('helga.plugins.poems.db')
def test_get_random_line_returns_none(db):
    db.haiku.find.return_value = db
    db.count.return_value = 0
    assert poems.get_random_line(5) is None


@patch('helga.plugins.poems.db')
def test_get_random_line_with_author(db):
    result = Mock()
    result.sort = result
    result.count.return_value = 1
    result.limit.return_value = result
    result.skip.return_value = result
    result.next.return_value = {'message': u'fives☃'}  # Ensure unicode test

    fake_find = Mock(return_value=result)

    db.haiku.find = fake_find
    poems.get_random_line(5, by='me')
    fake_find.assert_called_with({
        'syllables': 5,
        'author': {
            '$regex': re.compile('me', re.I)
        }
    })


@patch('helga.plugins.poems.db')
def test_get_random_line_with_search(db):
    result = Mock()
    result.sort = result
    result.count.return_value = 1
    result.limit.return_value = result
    result.skip.return_value = result
    result.next.return_value = {'message': 'fives1'}

    fake_find = Mock(return_value=result)

    db.haiku.find = fake_find
    poems.get_random_line(5, about='me')
    fake_find.assert_called_with({
        'syllables': 5,
        'message': {
            '$regex': re.compile('me', re.I)
        }
    })


@patch('helga.plugins.poems.db')
def test_get_random_line_with_search_or_author_retries(db):
    result = Mock()
    result.sort = result
    result.count.side_effect = (0, 1)
    result.limit.return_value = result
    result.skip.return_value = result
    result.next.return_value = {'message': 'fives1'}

    fake_find = Mock(return_value=result)

    db.haiku.find = fake_find
    poems.get_random_line(5, about='me')
    assert fake_find.call_count == 2
    assert fake_find.call_args_list[-1] == call({'syllables': 5})


@patch('helga.plugins.poems.add')
@patch('helga.plugins.poems.make_poem')
def test_use_fives(make_poem, add):
    make_poem.return_value = ['one', 'two', 'three']
    poem = poems.use(5, 'foo')
    assert 'foo' in (poem[0], poem[2])


@patch('helga.plugins.poems.add')
@patch('helga.plugins.poems.make_poem')
def test_use_fives_tanka(make_poem, add):
    make_poem.return_value = ['one', 'two', 'three', 'four', 'five']
    poem = poems.use(5, 'foo', poem_type='tanka')
    assert 'foo' in (poem[0], poem[2])


@patch('helga.plugins.poems.add')
@patch('helga.plugins.poems.make_poem')
def test_use_fives_does_not_duplicate(make_poem, add):
    make_poem.return_value = ['foo', 'two', 'three']
    poem = poems.use(5, 'foo')
    assert poem[0] == 'foo'
    assert poem[2] != 'foo'


@patch('helga.plugins.poems.add')
@patch('helga.plugins.poems.make_poem')
def test_use_sevens(make_poem, add):
    make_poem.return_value = ['one', 'two', 'three']
    poem = poems.use(7, 'foo')
    assert poem[1] == 'foo'


@patch('helga.plugins.poems.add')
@patch('helga.plugins.poems.make_poem')
def test_use_sevens_tanka(make_poem, add):
    make_poem.return_value = ['one', 'two', 'three', 'four', 'five']
    poem = poems.use(7, 'foo', poem_type='tanka')
    assert 'foo' in (poem[1], poem[3], poem[4])


@patch('helga.plugins.poems.get_random_line')
def test_make_poem(get_random_line):
    with patch.object(poems, 'fix_repitition', lambda s: s):
        # Poems have a random ordering
        with patch('helga.plugins.poems.random') as random:
            random.choice.side_effect = [0, 1]

            get_random_line.side_effect = ['one', 'two', 'three']
            assert poems.make_poem() == ['one', 'two', 'three']

            get_random_line.side_effect = ['one', 'two', 'three']
            assert poems.make_poem() == ['three', 'two', 'one']


@patch('helga.plugins.poems.get_random_line')
def test_make_poem_tanka(get_random_line):
    get_random_line.side_effect = ['one', 'two', 'three', 'four', 'five']
    poem = poems.make_poem(poem_type='tanka')

    # Poems have a random ordering
    assert poem in (
        ['one', 'two', 'three', 'four', 'five'],
        ['three', 'two', 'one', 'four', 'five'],
    )


@patch('helga.plugins.poems.get_random_line')
def test_make_poem_none_if_incomplete(get_random_line):
    get_random_line.side_effect = ['foo', 'bar', '']
    assert poems.make_poem(poem_type='haiku') is None


@patch('helga.plugins.poems.fix_repitition')
@patch('helga.plugins.poems.get_random_line')
def test_make_poem_adds_about_or_by_kwargs(get_random_line, fix_rep):
    poem = ['one', 'two', 'three']
    fix_rep.return_value = poem
    get_random_line.side_effect = poem
    poems.make_poem(about='me')
    args, kwargs = fix_rep.call_args
    assert sorted(args[0]) == sorted(poem)
    assert kwargs == {'about': 'me'}

    poem = ['one', 'two', 'three']
    get_random_line.side_effect = poem
    poems.make_poem(by='me')
    args, kwargs = fix_rep.call_args
    assert sorted(args[0]) == sorted(poem)
    assert kwargs == {'by': 'me'}


@patch('helga.plugins.poems.make_poem')
def test_poems_no_args_returns_poem(make_poem):
    poem = 'this is my poem'
    make_poem.return_value = poem
    assert poem == poems.poems(Mock(), '#bots', 'me', 'message', 'haiku', [])
    assert poem == poems.last_poem['#bots']
    make_poem.assert_called_with(poem_type='haiku')


@patch('helga.plugins.poems.make_poem')
def test_poems_about_or_by_subcommand(make_poem):
    args = [Mock(), '#bots', 'me', 'message', 'haiku']
    for extra in (['about', u'☃'], ['by', 'foo']):
        make_poem.reset_mock()
        poems.poems(*(args + [extra]))
        kwargs = {
            'poem_type': 'haiku',
            extra[0]: extra[1]
        }
        make_poem.assert_called_with(**kwargs)


@patch('helga.plugins.poems.blame')
def test_poems_blame_subcommand(blame):
    poems.poems(Mock(nickname='helga'), '#bots', 'me', 'message', 'haiku', ['blame'])
    blame.assert_called_with('#bots', requested_by='me', default_author='helga')


@patch('helga.plugins.poems.reactor')
def test_poems_tweet_subcommand(reactor):
    client = Mock()
    with pytest.raises(ResponseNotReady):
        poems.poems(client, '#bots', 'me', 'message', 'haiku', ['tweet'])
        reactor.assert_called_with(0, poems.tweet, client, '#bots', 'me')


@patch('helga.plugins.poems.add')
def test_poems_add_subcommand(add):
    subcmd_args = ['add', 'fives', 'unicode', u'☃']
    args = [Mock(), '#bots', 'me', 'message', 'haiku', subcmd_args]
    poems.poems(*args)
    add.assert_called_with(5, u'unicode ☃', 'me')


@patch('helga.plugins.poems.add_use')
def test_poems_add_use_subcommand(add_use):
    add_use.return_value = 'a poem'
    subcmd_args = ['add_use', 'fives', 'unicode', u'☃']
    args = [Mock(), '#bots', 'me', 'message', 'haiku', subcmd_args]
    poems.poems(*args)
    add_use.assert_called_with(5, u'unicode ☃', 'me')
    assert poems.last_poem['#bots'] == 'a poem'


@patch('helga.plugins.poems.use')
def test_poems_use_subcommand(use):
    use.return_value = 'a poem'
    subcmd_args = ['use', 'fives', 'unicode', u'☃']
    args = [Mock(), '#bots', 'me', 'message', 'haiku', subcmd_args]
    poems.poems(*args)
    use.assert_called_with(5, u'unicode ☃')
    assert poems.last_poem['#bots'] == 'a poem'


@patch('helga.plugins.poems.remove')
def test_poems_remove_subcommand(remove):
    subcmd_args = ['remove', 'fives', 'unicode', u'☃']
    args = [Mock(), '#bots', 'me', 'message', 'haiku', subcmd_args]
    poems.poems(*args)
    remove.assert_called_with(5, u'unicode ☃')


@patch('helga.plugins.poems.claim')
def test_poems_claim_subcommand(claim):
    subcmd_args = ['claim', 'fives', 'unicode', u'☃']
    args = [Mock(), '#bots', 'me', 'message', 'haiku', subcmd_args]
    poems.poems(*args)
    claim.assert_called_with(5, u'unicode ☃', 'me')


@patch('helga.plugins.poems.db')
def test_claim(db):
    resp = poems.claim(5, u'unicode snowman ☃', author='me')
    db.haiku.update.assert_called_with({'message': u'unicode snowman ☃'},
                                       {'$set': {'author': 'me'}})
    assert resp == u'me has claimed the line: unicode snowman ☃'


@patch('helga.plugins.poems.db')
def test_claim_unknown_line(db):
    db.haiku.update.side_effect = Exception
    resp = poems.claim(5, u'unicode snowman ☃', author='me')
    assert resp == "Sorry, I don't know that line."


def test_tweet_with_no_poem():
    client = Mock()
    poems.last_poem['#bots'] = None
    poems.tweet(client, '#bots', 'me')
    client.msg.assert_called_with('#bots', "me, why don't you try making one first")


@patch('helga.plugins.poems.send_tweet')
def test_tweet_with_failure(send_tweet):
    client = Mock()
    send_tweet.return_value = None
    poems.last_poem['#bots'] = ['foobar']
    poems.tweet(client, '#bots', 'me')
    client.msg.assert_called_with('#bots', "me, that probably did not work :(")


@patch('helga.plugins.poems.send_tweet')
def test_tweet(send_tweet):
    client = Mock()
    send_tweet.return_value = 'blah'
    poems.last_poem['#bots'] = ['foobar']
    poems.tweet(client, '#bots', 'me')
    client.msg.assert_called_with('#bots', 'blah')


def test_blame_with_no_poem():
    poems.last_poem['#bots'] = None
    resp = poems.blame('#bots', 'me')
    assert resp == "me, why don't you try making one first"


@patch('helga.plugins.poems.db')
def test_blame_uses_default_author(db):
    db.haiku.find_one.side_effect = Exception
    poems.last_poem['#bots'] = ['foo', 'bar', 'baz']
    resp = poems.blame('#bots', 'me', default_author='helga')
    assert resp == 'The last poem was brought to you by (in order): helga, helga, helga'


@patch('helga.plugins.poems.db')
def test_blame(db):
    poems.last_poem['#bots'] = ['foo', 'bar', 'baz']
    db.haiku.find_one.side_effect = [
        {'author': 'foo'},
        {'author': u'☃'},  # Test unicode
        {},  # No author
    ]
    resp = poems.blame('#bots', 'me', default_author='helga')
    assert resp == u'The last poem was brought to you by (in order): foo, ☃, helga'



# -*- coding: utf8 -*-
"""
Tests for helga.plugins
"""
from collections import defaultdict
from unittest import TestCase

from mock import Mock, patch
from pretend import stub

from helga import settings
from helga.plugins.core import (Command,
                                Match,
                                Plugin,
                                Registry,
                                ResponseNotReady,
                                command,
                                match,
                                preprocessor,
                                registry)


class RegistryTestCase(TestCase):

    def setUp(self):
        registry.plugins = {}
        registry.enabled_plugins = defaultdict(set)

        def foo():
            pass
        foo._plugins = []

        def bar():
            pass

        class Foo(Plugin):
            pass

        class Bar(object):
            pass

        self.valid_plugins = [foo, Foo, Foo()]
        self.invalid_plugins = [bar, Bar, Bar()]

        # For testing unicode compatibility
        self.snowman = u'☃'

    def test_prioritized(self):
        fake_plugin = stub(name='foo', priority=50)
        fake_decorated = stub(_plugins=[
            stub(name='bar', priority=10),
            stub(name='baz', priority=0),
            stub(name=self.snowman, priority=99),
        ])
        registry.plugins = {'foo': fake_plugin, 'bar': fake_decorated}
        registry.enabled_plugins['#bots'] = set(['foo', 'bar'])

        items = registry.prioritized('#bots')

        assert items[0].name == self.snowman
        assert items[1].name == 'foo'
        assert items[2].name == 'bar'
        assert items[3].name == 'baz'

    def test_prioritized_reversed(self):
        fake_plugin = stub(name='foo', priority=50)
        fake_decorated = stub(_plugins=[
            stub(name='bar', priority=10),
            stub(name='baz', priority=0),
            stub(name=self.snowman, priority=99),
        ])
        registry.plugins = {'foo': fake_plugin, 'bar': fake_decorated}
        registry.enabled_plugins['#bots'] = set(['foo', 'bar'])

        items = registry.prioritized('#bots', high_to_low=False)

        assert items[3].name == self.snowman
        assert items[2].name == 'foo'
        assert items[1].name == 'bar'
        assert items[0].name == 'baz'

    def test_prioritized_ignores_missing_plugin(self):
        fake_plugin = stub(name='foo', priority=50)
        registry.plugins = {'foo': fake_plugin}
        registry.enabled_plugins['#bots'] = set(['foo', 'bar'])

        items = registry.prioritized('#bots')

        assert len(items) == 1
        assert items[0].name == 'foo'

    def test_process_stops_when_async(self):
        things = [Mock(), Mock(), Mock()]

        # Make the middle one raise
        things[0].process.return_value = None
        things[1].process.side_effect = ResponseNotReady
        things[2].process.return_value = None

        with patch.object(registry, 'prioritized') as prio:
            prio.return_value = things
            assert [] == registry.process(None, '#bots', 'me', 'foobar')
            assert things[1].process.called
            assert not things[2].process.called

    def test_process_returns_all_responses(self):
        settings.PLUGIN_FIRST_RESPONDER_ONLY = False
        things = [Mock(), Mock(), Mock()]

        # Make the middle one raise
        things[0].process.return_value = ['foo', 'bar', None]
        things[1].process.return_value = self.snowman
        things[2].process.return_value = 'baz'

        with patch.object(registry, 'prioritized') as prio:
            prio.return_value = things
            response = registry.process(None, '#bots', 'me', 'foobar')
            assert response == [u'foo', u'bar', self.snowman, u'baz']

    def test_process_returns_first_response(self):
        settings.PLUGIN_FIRST_RESPONDER_ONLY = True
        things = [Mock(), Mock(), Mock()]

        # Make the middle one raise
        things[0].process.return_value = ['foo', 'bar', None]
        things[1].process.return_value = self.snowman
        things[2].process.return_value = 'baz'

        with patch.object(registry, 'prioritized') as prio:
            prio.return_value = things
            response = registry.process(None, '#bots', 'me', 'foobar')
            assert response == [u'foo', u'bar']

    def test_process_returns_unicode(self):
        plugin = Mock()
        plugin.process.return_value = ['foo', 'bar', self.snowman]

        with patch.object(registry, 'prioritized') as prio:
            prio.return_value = [plugin]
            responses = registry.process('', '', '', '')
            assert all(map(lambda x: isinstance(x, unicode), responses))

    def test_process_ignores_exception(self):
        settings.PLUGIN_FIRST_RESPONDER_ONLY = True
        things = [Mock(), Mock()]

        # Make the middle one raise
        things[0].process.side_effect = Exception
        things[1].process.return_value = 'foo'

        with patch.object(registry, 'prioritized') as prio:
            prio.return_value = things
            response = registry.process(None, '#bots', 'me', 'foobar')
            assert response == [u'foo']

    def test_registry_is_singleton(self):
        assert id(Registry()) == id(Registry())

    def test_register_raises_typeerror(self):
        for plugin in self.invalid_plugins:
            self.assertRaises(TypeError, registry.register, 'invalid', plugin)

    def test_register_valid_plugins(self):
        for plugin in self.valid_plugins:
            registry.register(repr(plugin), plugin)
            assert repr(plugin) in registry.plugins

    def test_register_plugin_handles_unicode(self):
        for plugin in self.invalid_plugins:
            self.assertRaises(TypeError, registry.register, 'invalid', plugin)

        for plugin in self.valid_plugins:
            name = u'{0}-{1}'.format(self.snowman, repr(plugin))
            registry.register(name, plugin)
            assert name in registry.plugins

    def test_all_plugins(self):
        registry.register('foo', self.valid_plugins[0])
        assert set(['foo']) == registry.all_plugins

    def test_get_plugin(self):
        registry.register('foo', self.valid_plugins[0])
        registry.register(self.snowman, self.valid_plugins[0])
        assert self.valid_plugins[0] == registry.get_plugin('foo')
        assert self.valid_plugins[0] == registry.get_plugin(self.snowman)

    def test_get_plugin_returns_none(self):
        assert registry.get_plugin('foo') is None

    def test_enable(self):
        assert 'foo' not in registry.enabled_plugins['#foo']
        assert self.snowman not in registry.enabled_plugins['#foo']
        registry.enable('#foo', 'foo', self.snowman)
        assert 'foo' in registry.enabled_plugins['#foo']
        assert self.snowman in registry.enabled_plugins['#foo']

    def test_disable(self):
        registry.enable('#foo', 'foo', self.snowman)
        assert 'foo' in registry.enabled_plugins['#foo']
        assert self.snowman in registry.enabled_plugins['#foo']
        registry.disable('#foo', 'foo', self.snowman)
        assert 'foo' not in registry.enabled_plugins['#foo']
        assert self.snowman not in registry.enabled_plugins['#foo']

    @patch('helga.plugins.core.pkg_resources')
    @patch('helga.plugins.core.smokesignal')
    def test_load(self, signal, pkg_resources):
        entry_points = [
            Mock(load=lambda: 'foo'),
            Mock(load=lambda: 'snowman'),
            Mock(),
        ]
        entry_points[0].name = 'foo'
        entry_points[1].name = self.snowman

        # Exceptions should not bomb the load process
        entry_points[2].name = 'bar'
        entry_points[2].load.side_effect = Exception

        pkg_resources.iter_entry_points.return_value = entry_points

        with patch.object(registry, 'register') as register:
            registry.load()
            assert ('foo', 'foo') == register.call_args_list[0][0]
            assert (self.snowman, 'snowman') == register.call_args_list[1][0]
            assert len(register.call_args_list) == 2  # Only the first two

        # Ensure that we sent the signal
        signal.emit.assert_called_with('plugins_loaded')

    def test_reload_not_registered(self):
        assert 'Unknown plugin' in registry.reload('foo')
        assert 'Unknown plugin' in registry.reload(self.snowman)

    @patch('helga.plugins.core.pkg_resources')
    @patch('helga.plugins.core.sys')
    @patch('__builtin__.reload')
    def test_reload(self, reloader, sys, pkg_resources):
        entry_points = [
            Mock(module_name='foo', load=lambda: 'loaded'),
            Mock(module_name='snowman', load=lambda: 'loaded')
        ]
        entry_points[0].name = 'foo'
        entry_points[1].name = self.snowman
        pkg_resources.iter_entry_points.return_value = entry_points

        sys.modules = {
            'foo': entry_points[0],
            'snowman': entry_points[1],
        }

        registry.plugins = ['foo', self.snowman]

        for name in ('foo', self.snowman):
            with patch.object(registry, 'register') as register:
                assert registry.reload(name)
                register.assert_called_with(name, 'loaded')

    @patch('helga.plugins.core.pkg_resources')
    @patch('helga.plugins.core.sys')
    @patch('__builtin__.reload')
    def test_reload_returns_false_on_exception(self, reloader, sys, pkg_resources):
        module = Mock(module_name='foo')
        module.name = 'foo'
        module.load.side_effect = Exception

        entry_points = [module]
        pkg_resources.iter_entry_points.return_value = entry_points
        sys.modules = {'foo': module}
        registry.plugins = ['foo']

        with patch.object(registry, 'register') as register:
            assert not registry.reload('foo')
            assert not register.called

    def test_preprocess(self):
        plugins = [Mock(), Mock(), Mock()]

        # Raising an exception shouldn't affect the others
        plugins[0].preprocess.return_value = ('foo', 'bar', self.snowman)
        plugins[1].preprocess.side_effect = Exception
        plugins[2].preprocess.return_value = ('abc', 'def', 'ghi')

        with patch.object(registry, 'prioritized') as prio:
            prio.return_value = plugins

            # Return value is what the last plugin returns
            retval = registry.preprocess(None, '#bots', 'me', self.snowman)
            assert retval == ('abc', 'def', 'ghi')

            # Should receive what the first plugin changed
            plugins[2].preprocess.assert_called_with(None, 'foo', 'bar', self.snowman)

            # Exception raising preprocess should have at least been called
            assert plugins[1].preprocess.called


class PluginTestCase(TestCase):

    def setUp(self):
        self.plugin = Plugin()
        self.client = Mock(nickname='helga')

    def test_preprocessor_decorator(self):
        @preprocessor
        def foo(client, channel, nick, message):
            return 'foo', 'bar', 'baz'

        expected = ('foo', 'bar', 'baz')
        args = (self.client, '#bots', 'me', 'foobar')

        assert hasattr(foo, '_plugins')
        assert len(foo._plugins) == 1
        assert expected == foo(*args)
        assert expected == foo._plugins[0].preprocess(*args)

    def test_preprocess_with_priority(self):
        @preprocessor(10)
        def foo(client, channel, nick, message):
            return 'foo', 'bar', 'baz'

        expected = ('foo', 'bar', 'baz')
        args = (self.client, '#bots', 'me', 'foobar')

        assert hasattr(foo, '_plugins')
        assert len(foo._plugins) == 1
        assert expected == foo(*args)
        assert expected == foo._plugins[0].preprocess(*args)
        assert 10 == foo._plugins[0].priority

    def test_base_plugin_process_calls_run(self):
        plugin = Plugin()
        with patch.object(plugin, 'run') as run:
            plugin.process('foo', 'bar', 'baz', 'qux')
            run.assert_called_with('foo', 'bar', 'baz', 'qux')


class CommandTestCase(TestCase):

    def setUp(self):
        self.cmd = Command('foo', aliases=('bar', 'baz'), help='foo cmd')
        self.client = Mock(nickname='helga')

    def test_init_does_not_overwrite_things(self):
        class MyCommand(Command):
            command = 'dothis'
            aliases = ('foo', 'bar', 'baz', 'f')
            help = 'my command'

        cmd = MyCommand()
        assert cmd.command == 'dothis'
        assert cmd.aliases == ('foo', 'bar', 'baz', 'f')
        assert cmd.help == 'my command'

    def test_parse_handles_main_command(self):
        assert 'foo' == self.cmd.parse('helga', 'helga foo')[0]

    @patch('helga.plugins.core.settings')
    def test_parse_handles_char_prefix(self, settings):
        settings.COMMAND_PREFIX_CHAR = '#'
        assert 'foo' == self.cmd.parse('helga', '#foo')[0]

    def test_parse_handles_aliases(self):
        assert 'bar' == self.cmd.parse('helga', 'helga bar')[0]
        assert 'baz' == self.cmd.parse('helga', 'helga baz')[0]

    def test_parse_with_punctuation(self):
        assert 'foo' == self.cmd.parse('helga', 'helga: foo')[0]
        assert 'foo' == self.cmd.parse('helga', 'helga, foo')[0]
        assert 'foo' == self.cmd.parse('helga', 'helga ----> foo')[0]

    def test_parse_does_not_handle(self):
        assert '' == self.cmd.parse('helga', 'helga qux')[0]

    def test_parse_returns_args(self):
        assert ['1', '2', '3'] == self.cmd.parse('helga', 'helga foo 1 2 3')[1]

    def test_parse_handles_longest_command_first(self):
        with patch.object(self.cmd, 'aliases', ['b', 'bar']):
            for check in ('b', 'bar'):
                cmd, args = self.cmd.parse('helga', 'helga %s baz' % check)
                assert cmd == check
                assert args == ['baz']

    def test_parse_does_not_handle_something_else(self):
        assert ('', []) == self.cmd.parse('helga', 'helga fun')

    def test_parse_handles_unicode(self):
        snowman = u'☃'
        disapproval = u'ಠ_ಠ'
        cmd = u'helga {0} {1}'.format(snowman, disapproval)
        self.cmd.command = snowman
        assert (snowman, [disapproval]) == self.cmd.parse('helga', cmd)

    def test_process_for_different_command_returns_none(self):
        assert self.cmd.process(self.client, '#bots', 'me', 'helga qux') is None

    def test_process_calls_class_run_method(self):
        self.cmd.run = lambda client, chan, nick, msg, cmd, args: 'run'
        assert 'run' == self.cmd.process(self.client, '#bots', 'me', 'helga foo')

    def test_process_calls_custom_runner(self):
        cmd = Command('foo', aliases=('bar', 'baz'), help='foo command')
        cmd.run = lambda client, chan, nick, msg, cmd, args: 'runner'
        assert 'runner' == cmd.process(self.client, '#bots', 'me', 'helga foo')

    def test_multiple_decorators(self):
        @command('foo')
        @command('bar')
        def foobar(*args):
            return args[-2]

        assert len(foobar._plugins) == 2
        assert 'bar' == foobar._plugins[0](self.client, '#bots', 'me', 'helga bar')
        assert 'foo' == foobar._plugins[1](self.client, '#bots', 'me', 'helga foo')

    def test_decorator_using_command(self):
        @command('foo')
        def foo(client, chan, nick, msg, cmd, args):
            return 'bar'

        assert 'bar' == foo._plugins[0](self.client, '#bots', 'me', 'helga foo')

    def test_decorator_using_alias(self):
        @command('foo', aliases=['baz'])
        def foo(client, chan, nick, msg, cmd, args):
            return 'bar'

        assert 'bar' == foo._plugins[0](self.client, '#bots', 'me', 'helga baz')


class MatchTestCase(TestCase):

    def setUp(self):
        self.match = Match('foo')
        self.client = Mock()

    def test_init_does_not_overwrite_things(self):
        class MyMatch(Match):
            pattern = 'foo'

        m = MyMatch()
        assert m.pattern == 'foo'

    def test_match_using_callable(self):
        self.match.pattern = lambda m: 'foobar'
        assert 'foobar' == self.match.match('this is a foo message')

    def test_match_using_simple_pattern(self):
        self.match.pattern = r'foo-(\d+)'
        assert ['123'] == self.match.match('this is about foo-123')

    def test_match_returns_none_on_typeerror(self):
        self.match.pattern = Mock(side_effect=TypeError)
        assert self.match.match('this is a foo message') is None

    def test_simple_decorator(self):
        @match('foo-(\d+)')
        def foo(client, chan, nick, msg, matches):
            return matches[0]

        assert '123' == foo._plugins[0](self.client, '#bots', 'me', 'this is about foo-123')

    def test_callable_decorator(self):
        @match(lambda x: x.startswith('foo'))
        def foo(client, chan, nick, msg, matches):
            return 'bar'

        assert 'bar' == foo._plugins[0](self.client, '#bots', 'me', 'foo at the start')
        assert foo._plugins[0](self.client, '#bots', 'me', 'not at the start foo') is None

    def test_match_with_unicode(self):
        @match(u'☃')
        def snowman_match(client, chan, nick, msg, matches):
            return 'snowman'
        assert 'snowman' == snowman_match._plugins[0](self.client, '#bots', 'me', u'☃')



import functools
import pkg_resources
import re

from twisted.internet import reactor
from twisted.web import server, resource

import smokesignal

from helga import log, settings
from helga.plugins import Command, registry


logger = log.getLogger(__name__)


class WebhookPlugin(Command):
    """
    Webhooks provides a way to expose HTTP endpoints that can interact with
    helga. An HTTP listener will be started on a port specified by setting
    WEBHOOKS_PORT (default 8080). An additional, optional setting that can
    be used for routes requiring HTTP basic auth is WEBOOKS_CREDENTIALS, which
    should be a list of tuples, where each tuple is a pair of (username, password).

    Routes are HTTP path endpoints. On the surface they are just python
    callables decorated using @route. The route decorate (described below)
    must be given a path regex, and optional HTTP methods to accept. Webhook
    plugins must be registered in the same way normal plugins are registered,
    using setuptools entry_points. However, they must belong to the entry_point
    group ``helga_webhooks``. For example::

        setup(entry_points={
            'helga_webhooks': [
                'api = myapi.decorated_route'
            ]
        })

    This will make new webhooks installable in a very simple, and repeatable way.

    The exposed webhook plugin gives control for starting and stopping the HTTP
    service, as well as listing available routes.
    """
    command = 'webhooks'
    help = ('HTTP service for interacting with helga. Command options usage: '
            'helga webhooks (start|stop|routes). Note: start/stop'
            'can be run only by helga operators')

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)

        self.root = None
        self.site = None
        self.port = getattr(settings, 'WEBHOOKS_PORT', 8080)

        @smokesignal.on('signon')
        def setup(client):  # pragma: no cover
            self._start(client)
            self._init_routes()

    def _init_routes(self):
        for entry_point in pkg_resources.iter_entry_points(group='helga_webhooks'):
            try:
                logger.debug('Loading webhook %s', entry_point.name)
                entry_point.load()
            except:  # pragma: no cover
                logger.exception('Error loading webhook %s', entry_point)

    def _start(self, client=None):
        logger.info('Starting webhooks service on port %s', self.port)

        if self.root is None:
            self.root = WebhookRoot(client)
        else:
            self.root.irc_client = client

        if self.site is None:
            self.site = server.Site(self.root)

        self.tcp = reactor.listenTCP(self.port, self.site)

    def _stop(self):
        logger.info('Stopping webhooks service on port %s', self.port)
        self.tcp.stopListening()
        self.tcp.loseConnection()
        self.tcp = None

    def add_route(self, fn, path, methods):
        """
        Adds a route to the root web resource
        """
        self.root.add_route(fn, path, methods)  # pragma: no cover

    def list_routes(self, client, nick):
        """
        Messages a user with all webhook routes and their supported HTTP methods
        """
        client.msg(nick, u'{0}, here are the routes I know about'.format(nick))
        for pattern, route in self.root.routes.iteritems():
            http_methods = route[0]  # Route is a tuple (http_methods, function)
            client.msg(nick, u'[{0}] {1}'.format(','.join(http_methods), pattern))

    def control(self, action):
        """
        Operator-level control over stop/start of the running TCP listener
        """
        running = self.tcp is not None

        if action == 'stop':
            if running:
                self._stop()
                return u"Webhooks service stopped"
            return u"Webhooks service not running"

        if action == 'start':
            if not running:
                self._start()
                return u"Webhooks service started"
            return u"Webhooks service already running"

    def run(self, client, channel, nick, msg, cmd, args):
        try:
            subcmd = args[0]
        except IndexError:
            subcmd = 'routes'

        if subcmd == 'routes':
            client.me(channel, u'whispers to {0}'.format(nick))
            self.list_routes(client, nick)
        elif subcmd in ('start', 'stop'):
            if nick not in client.operators:
                return u"Sorry {0}, Only an operator can do that".format(nick)
            return self.control(subcmd)


class WebhookRoot(resource.Resource):
    isLeaf = True

    def __init__(self, irc_client, *args, **kwargs):
        self.irc_client = irc_client

        # Routes is a dict of regex path -> (allowed_methods_list, callable)
        self.routes = {}

    def add_route(self, fn, path, methods):
        """
        Adds a webhook route for service
        """
        self.routes[path] = (methods, fn)

    def render(self, request):
        """
        Handle finding and dispatching the route matching the incoming request path
        """
        request.setHeader('Server', 'helga')
        for pat, route in self.routes.iteritems():
            match = re.match(pat, request.path)
            if match:
                break
        else:
            request.setResponseCode(404)
            return u'404 Not Found'

        # Ensure that this route handles the request method
        methods, fn = route
        if request.method.upper() not in methods:
            request.setResponseCode(405)
            return u'405 Method Not Allowed'

        return fn(request, self.irc_client, **match.groupdict())


def authenticated(fn):
    """
    A wrapper for a webhook route that ensures value HTTP basic auth
    credentials are supplied. Example::

        @authenticated
        @route('/foo/bar')
        def my_endpoint(request, irc_client):
            pass

    Any valid credentials in the setting WEBHOOKS_CREDENTIALS, which should
    be a list of tuples (username, password) will be allowed. If no valid
    credentials are supplied, an HTTP 401 is returned
    """
    @functools.wraps(fn)
    def ensure_authenticated(request, *args, **kwargs):
        for user, password in getattr(settings, 'WEBHOOKS_CREDENTIALS', []):
            if user == request.getUser() and password == request.getPassword():
                return fn(request, *args, **kwargs)

        # No valid basic auth provided
        request.setResponseCode(401)
        return u'401 Unauthorized'
    return ensure_authenticated


def route(path, methods=None):
    """
    Decorator to register a webhook route. You must specify a path
    regular expression, and optionally a list of HTTP methods to accept.
    If you do not specify a list of HTTP methods, only GET requests will
    be served. All regex matches must be named groups and they will be passed
    as keyword arguments. For those who are familiar with something
    like Flask, this will feel very similar. Example::

        @route(r'/users/(?P<user_id>[0-9]+)')
        def get_user(request, irc_client, user_id):
            pass

        @route(r'/foo/(?P<subpage>[a-z]+)/(?P<page_id>[0-9]+)', methods=['GET', 'POST'])
        def page(request, irc_client, subpage, page_id):
            pass

    Note that route callables (shown above) must accept as the first two
    positional arguments a request object, and the current irc client.
    """
    plugin = registry.get_plugin('webhooks')
    if methods is None:
        methods = ['GET']

    def wrapper(fn):
        if plugin is not None:
            plugin.add_route(fn, path, methods)
        return fn

    return wrapper



import random

from itertools import ifilter

import smokesignal

from helga import log
from helga.db import db
from helga.plugins import command, ACKS
from helga.plugins.core import registry


logger = log.getLogger(__name__)


@smokesignal.on('signon')
def auto_enable_plugins(*args):
    pred = lambda rec: rec['plugin'] in registry.all_plugins

    for rec in ifilter(pred, db.auto_enabled_plugins.find()):
        for channel in rec['channels']:
            logger.info('Auto-enabling plugin %s on channel %s', rec['plugin'], channel)
            registry.enable(channel, rec['plugin'])


def list_plugins(client, channel):
    enabled = set(registry.enabled_plugins[channel])
    available = registry.all_plugins - enabled

    return [
        u'Plugins enabled on this channel: {0}'.format(', '.join(sorted(enabled))),
        u'Available plugins: {0}'.format(', '.join(sorted(available))),
    ]


def _filter_valid(channel, *plugins):
    return filter(lambda p: p in registry.all_plugins, plugins)


def enable_plugins(client, channel, *plugins):
    valid_plugins = _filter_valid(channel, *plugins)
    if not valid_plugins:
        return u"Sorry, but I don't know about these plugins: {0}".format(', '.join(plugins))

    registry.enable(channel, *valid_plugins)

    for p in valid_plugins:
        rec = db.auto_enabled_plugins.find_one({'plugin': p})
        if rec is None:
            db.auto_enabled_plugins.insert({'plugin': p, 'channels': [channel]})
        elif channel not in rec['channels']:
            rec['channels'].append(channel)
            db.auto_enabled_plugins.save(rec)

    return random.choice(ACKS)


def disable_plugins(client, channel, *plugins):
    valid_plugins = _filter_valid(channel, *plugins)
    if not valid_plugins:
        return u"Sorry, but I don't know about these plugins: {0}".format(', '.join(plugins))

    registry.disable(channel, *valid_plugins)

    for p in valid_plugins:
        rec = db.auto_enabled_plugins.find_one({'plugin': p})
        if rec is None or channel not in rec['channels']:
            continue

        rec['channels'].remove(channel)
        db.auto_enabled_plugins.save(rec)

    return random.choice(ACKS)


@command('plugins', help="Plugin management. Usage: helga plugins (list|(enable|disable) (<name> ...))")
def manager(client, channel, nick, message, cmd, args):
    """
    Manages listing plugins, or enabling and disabling them
    """
    subcmd = args[0]

    if subcmd == 'list':
        return list_plugins(client, channel)

    if subcmd == 'enable':
        return enable_plugins(client, channel, *args[1:])

    if subcmd == 'disable':
        return disable_plugins(client, channel, *args[1:])



import functools
import pkg_resources
import random
import re
import sys

from collections import defaultdict
from itertools import ifilter, imap

import smokesignal

from helga import log, settings
from helga.util.encodings import to_unicode


logger = log.getLogger(__name__)


ACKS = [
    'roger',
    '10-4',
    'no problem',
    'will do',
    'you got it',
    'anything you say',
    'sure thing',
    'ok',
    'right-o',
    'consider it done',
]


PRIORITY_LOW = 25
PRIORITY_NORMAL = 50
PRIORITY_HIGH = 75


def random_ack():
    return random.choice(ACKS)


class ResponseNotReady(Exception):
    """
    Exception raised by plugins that perform some async operation using
    twisted deferreds. If the bot is configured to only allow the first plugin
    response (by default), then any plugin raising this will prevent further
    plugin execution
    """


class Registry(object):
    """
    Simple plugin registry that handles dispatching messages to registered plugins.
    Plugins can be disabled per channel. By default, plugins are loaded, but not
    enabled unless ENABLED_PLUGINS is set in helga settings. This is done so that
    potentially annoying plugins can be enabled on-demand
    """
    __instance = None

    def __new__(cls, *args, **kwargs):
        """
        Only here so we only maintain one registry for the life of the application. There
        is some state-specific things that shouldn't be lost in the event the IRC client
        loses the connection to the server
        """
        if cls.__instance is None:
            cls.__instance = super(Registry, cls).__new__(cls, *args, **kwargs)
        return cls.__instance

    def __init__(self):
        if not hasattr(self, 'plugins'):
            self.plugins = {}

        if not hasattr(self, 'enabled_plugins'):
            # Enabled plugins is a dict: channel -> set()
            self.enabled_plugins = defaultdict(lambda: set(getattr(settings, 'ENABLED_PLUGINS', [])))

        smokesignal.on('started', self.load)

    def register(self, name, fn_or_cls):
        # Make sure we're working with an instance
        try:
            if issubclass(fn_or_cls, Plugin):
                fn_or_cls = fn_or_cls()
        except TypeError:
            pass

        if not (isinstance(fn_or_cls, Plugin) or hasattr(fn_or_cls, '_plugins')):
            raise TypeError(u"Plugin {0} must be a subclass of Plugin, or a decorated function".format(name))

        logger.info('Registered plugin %s', name)
        self.plugins[name] = fn_or_cls

    @property
    def all_plugins(self):
        return set(self.plugins.keys())

    def get_plugin(self, name):
        return self.plugins.get(name, None)

    def disable(self, channel, *plugins):
        self.enabled_plugins[channel] = self.enabled_plugins[channel].difference(set(plugins))

    def enable(self, channel, *plugins):
        self.enabled_plugins[channel] = self.enabled_plugins[channel].union(set(plugins))

    def load(self):
        """
        Locate all setuptools entry points by the name 'helga_plugins'
        and initialize them. Any third-party library may register a plugin by
        addint the following to their setup.py::

            entry_points = {
                'helga_plugins': [
                    'plugin_name = mylib.mymodule:MyPluginClass',
                ],
            },

        """
        for entry_point in pkg_resources.iter_entry_points(group='helga_plugins'):
            try:
                logger.debug('Loading entry_point %s', entry_point.name)
                self.register(entry_point.name, entry_point.load())
            except:
                logger.exception('Error initializing plugin %s', entry_point)
        smokesignal.emit('plugins_loaded')

    def reload(self, name):
        if name not in self.plugins:
            return u"Unknown plugin '{0}'. Is it installed?".format(name)

        for entry_point in pkg_resources.iter_entry_points(group='helga_plugins'):
            if entry_point.name != name:
                continue

            try:
                reload(sys.modules[entry_point.module_name])
                self.register(entry_point.name, entry_point.load())
                return True
            except:
                logger.exception('Failed to reload plugin %s', entry_point)
                return False

    def prioritized(self, channel, high_to_low=True):
        """
        Gets a list of plugins in order according to their priority
        """
        plugins = []
        for name in self.enabled_plugins[channel]:
            if name not in self.plugins:
                logger.debug('Plugin %s may not be installed or have incorrect entry_point information', name)
                continue

            # Decorated functions will have this
            if hasattr(self.plugins[name], '_plugins'):
                plugins.extend(self.plugins[name]._plugins)
            else:
                plugins.append(self.plugins[name])

        return sorted(plugins, key=lambda p: getattr(p, 'priority', PRIORITY_NORMAL), reverse=high_to_low)

    def preprocess(self, client, channel, nick, message):
        for plugin in self.prioritized(channel):
            try:
                channel, nick, message = plugin.preprocess(client, channel, nick, message)
            except:
                logger.exception('Calling preprocess on plugin %s failed', plugin)
                continue

        return channel, nick, message

    def process(self, client, channel, nick, message):
        responses = []
        first_responder = getattr(settings, 'PLUGIN_FIRST_RESPONDER_ONLY', False)

        for plugin in self.prioritized(channel):
            try:
                resp = plugin.process(client, channel, nick, message)
            except ResponseNotReady:
                break
            except:
                logger.exception('Calling process on plugin %s failed', plugin)
                continue

            if not resp:
                continue

            # Chained decorator style plugins return a list of strings
            if isinstance(resp, (tuple, list)):
                # Be sure to filter Nones, then strip
                responses.extend(imap(lambda s: (s or '').strip(), resp))
            else:
                responses.append(resp.strip())

            if responses and first_responder:
                break

        # FIXME: Explicit conversion to unicode might not make sense. Perpahs
        # a warning should be sent to the user? Or do we even care?
        return map(to_unicode, ifilter(bool, responses))


registry = Registry()


class Plugin(object):
    """
    Base class for creating helga plugins. Plugins have a minimal API, and there
    are essentially two methods that require some implementation, ``__call__`` and
    ``run``. Although these methods may seem to have a similar meaning, they both
    have different purposes. An implementation for ``__call__`` is required by helga's
    plugin registry, since it expects plugins to be callable. The call arguments are
    expected to be consistent for all plugin implementations. On the other hand, the
    ``run`` method is used to do the actual legwork for a plugin. At a minimum, this
    should accept the same arguments as ``__call__`` but subclasses my define any additional
    arguments if necessary.

    In other words, ``process`` should decide if a message is worthy of a response, and
    ``run`` should provide the response. As a simple example::

        import time

        class MyPlugin(Plugin):
            def run(self, channel, nick, message):
                return 'Current timestamp: {0}'.format(time.time())

            def process(self, channel, nick, message):
                if message.startswith('!time'):
                    return self.run(channel, nick, message)
    """
    priority = PRIORITY_NORMAL

    def __init__(self, priority=PRIORITY_NORMAL):
        self.priority = priority

    def run(self, client, channel, nick, message, *args, **kwargs):
        """
        Runs the plugin to generate a response. At a minimum this should accept
        three arguments. It should also either return None, if no response is to be
        sent back over IRC, or a non-empty string.

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        """
        return None  # pragma: no cover

    def preprocess(self, client, channel, nick, message):
        """
        A preprocessing filter for plugins. This allows a plugin to modify a received
        message prior to be being sent to a plugin's ``process`` method. Implementations
        of this must return a three-tuple containing, any of which could be modified:

        - **channel**: the channel on which the message was received
        - **nick**: the current nick of the message sender
        - **message**: the message string itself

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        """
        return channel, nick, message  # pragma: no cover

    def process(self, client, channel, nick, message):
        """
        This is the global entry point for plugins according to helga's plugin registry.
        Each plugin will be called when a message is received over IRC with the channel
        in which the message was received, the nick of the user that sent the message, and
        the message itself. Much like ``run``, this should either return None, if no
        response is to be sent back over IRC, or a non-empty string. In most cases, this
        should just return whatever the return value of calling ``self.run`` is.

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        """
        return self.run(client, channel, nick, message)

    def decorate(self, fn, preprocessor=False):
        """
        A helper used for establishing decorators for plugin classes. This does nothing
        more than monkey patch the ``run`` method of the instance with whatever function
        is being decorated. Note that the decorated function should accept whatever arguments
        the subclass implementation sends to its ``run`` method. Instances of plugin objects
        are kept in a list attribute of the decorated function. This allows chainable decorators
        that function as intended, providing improved plugin functionality.
        """
        if preprocessor:
            self.preprocess = fn
        else:
            self.run = fn

        try:
            fn._plugins.append(self)
        except AttributeError:
            fn._plugins = [self]

        return fn

    def __call__(self, client, channel, nick, message):
        """
        Proxy for ``process``
        """
        return self.process(client, channel, nick, message)


class Command(Plugin):
    """
    A command is a type of plugin which requries a user to specifically ask
    for some action. This can be something simple like "helga haiku" or
    "helga google airspeed velocity of an unladen swallow".

    Command subclasses must provide three class-level attributes to function
    properly:

    - **command**: The command string that invokes this action
    - **aliases**: An optional list of string aliases that can also invoke this command
    - **help**: Useful help string on how to use this command

    In addition to these attributes, subclasses must also implement a ``run`` method
    that accepts six arguments:

    - **client**: an instance of :class:`helga.comm.Client`
    - **channel**: the channel on which the message was received
    - **nick**: the current nick of the message sender
    - **message**: the message string itself
    - **command**: the parsed command, which is either the preferred command or a command alias
    - **args**: a list of strings, space delimited, that follow the command

    The following is a simple example::

        class Haiku(Command):
            command = 'foo'
            aliases = ('bar',)
            help = 'Responds to `helga foo` with the string bar'

            def run(self, channel, nick, message, command, args):
                return 'bar'

    Using the above example may produce the following results in IRC::

        <sduncan> helga foo
        <helga> sduncan said foo
        <sduncan> helga bar
        <helga> sduncan said bar
    """

    command = ''
    aliases = []
    help = ''

    def __init__(self, command='', aliases=None, help='', priority=PRIORITY_NORMAL):
        super(Command, self).__init__(priority)
        self.command = command or self.command
        self.aliases = aliases or self.aliases
        self.help = help or self.help

    def parse(self, botnick, message):
        """
        Parse the incoming message using the current nick of the bot, the defined
        command string of this object, plus any aliases. Will return the actual
        command parsed (which could be an alias), plus whitespaced delimited list
        of strings that follow the parsed command.

        :param botnick: helga's current nickname
        :param message: the incoming IRC message
        :returns: string of parsed command, whitespaced delimited string list of args
        """
        choices = [self.command] + list(self.aliases)

        # Sort choices from longest to shortest. This will ease a quirk where
        # short alias versions will trump the more verbose ones
        choices = sorted(choices, key=len, reverse=True)

        nick_prefix = ''

        # Handle multiple ways to parse this command
        if getattr(settings, 'COMMAND_PREFIX_BOTNICK', True):
            nick_prefix = '{0}\W*\s'.format(botnick)

        prefixes = filter(bool, [nick_prefix, getattr(settings, 'COMMAND_PREFIX_CHAR', '!')])
        prefix = '({0})'.format('|'.join(prefixes))

        pat = ur'^{0}({1})($|\s(.*)$)'.format(prefix, '|'.join(choices))

        try:
            _, cmd, _, argstr = re.findall(pat, message, re.IGNORECASE)[0]
        except (IndexError, ValueError):
            # FIXME: Log here?
            return u'', []

        return cmd, filter(bool, argstr.strip().split(' '))

    def run(self, client, channel, nick, message, command, args):
        """
        This is where any actual work should be done for a command. Generally, this
        is only used for subclasses of the base Command class. The @command decorator
        will monkey patch this method with whatever the decorated function is.

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        :param command: the parsed command, which is either the preferred command or a command alias
        :param args: a list of strings, space delimited, that follow the command
        """
        return None  # pragma: no cover

    def process(self, client, channel, nick, message):
        """
        Processes a message sent by a user on a given channel. This will return
        None if this command should respond to the incoming message, or if there
        is an exception raised from calling the ``run`` method or attribute of the
        instance.

        Generally, subclasses should not have to worry about this method, and
        instead, should focus on the implementation of ``run``, which is called should
        this command be invoked.

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        """
        command, args = self.parse(client.nickname, message)
        if command != self.command and command not in self.aliases:
            return None

        return self.run(client, channel, nick, message, command, args)


class Match(Plugin):
    """
    A match is a type of plugin that will make helga respond if the contents
    of a user's message match a pattern. For example, if a match plugin looks
    for the string "foo", helga will respond with "bar" without being asked
    to do so. Subclasses must provide one class-level attribute:

    - **pattern**: either a valid regex pattern string, or a callable

    If the match pattern is a callable, it must accept a single string argument,
    which will be the IRC message to check, and it must return a value that can be
    evaluated as either True or False via bool(). Good practice here is to
    return the list of strings that were matched or found, but this is not
    required.

    In addition to this class attribute, subclasses must also implement a ``run``
    method that accepts five arguments:

    - **client**: an instance of :class:`helga.comm.Client`
    - **channel**: the channel on which the message was received
    - **nick**: the current nick of the message sender
    - **message**: the message string itself
    - **matches**: if the ``pattern`` attribute is a callable, this will be its return value,
      otherwise it will be the return value of ``re.findall``.

    The following is a simple example::

        class FooMatch(Match):
            pattern = r'foo-(\d+)'

            def run(self, channel, nick, message, matches):
                return '{0} is talking about foo thing {0}'.format(nick, matches[0])

    Using the above example may produce the following results in IRC::

        <sduncan> this is about foo-123
        <helga> sduncan is talking about foo thing 123
    """
    pattern = ''

    def __init__(self, pattern='', priority=PRIORITY_LOW):
        super(Match, self).__init__(priority)
        self.pattern = pattern or self.pattern

    def run(self, client, channel, nick, message, matches):
        """
        This is where any actual work should be done for a command. Generally, this
        is only used for subclasses of the base Command class. The @command decorator
        will monkey patch this method with whatever the decorated function is.

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        :param matches: if the ``pattern`` attribute of this class is a callable, this will be its
                        return value, otherwise it will be the return value of ``re.findall``
        """
        return None  # pragma: no cover

    def match(self, message):
        """
        Matches a message against the pattern defined for this class. If the pattern attribute
        is a callable, it is called with the message as its only argument and that value is returned.
        Otherwise, the pattern attribute is used as a regular expression string argument to
        ``re.findall`` and that value is returned.

        :param message: the message received over IRC
        """
        if callable(self.pattern):
            fn = self.pattern
        else:
            fn = functools.partial(re.findall, self.pattern)

        try:
            return fn(message)
        except TypeError:
            return None

    def process(self, client, channel, nick, message):
        """
        Processes a message sent by a user on a given channel. This will return
        None if the message does not match the plugin's pattern, or a non-empty string
        (the return value of ``run``) if it does match.

        Generally, subclasses should not have to worry about this method, and
        instead, should focus on the implementation of ``run``, which is called should
        the incoming message be matched against the plugin's pattern.

        :param client: an instance of :class:`helga.comm.Client`
        :param channel: the channel on which the message was received
        :param nick: the current nick of the message sender
        :param message: the message string itself
        """
        matches = self.match(message)
        if not bool(matches):
            return None

        return self.run(client, channel, nick, message, matches)


def command(command, aliases=None, help='', priority=PRIORITY_NORMAL):
    """
    A decorator for creating simple commands where subclassing the ``Command``
    plugin type may be overkill. This is generally the easiest way to create helga
    commands. However, decorated functions must adhere to accepting six arguments:

    - **client**: an instance of :class:`helga.comm.Client`
    - **channel**: the channel on which the message was received
    - **nick**: the current nick of the message sender
    - **message**: the message string itself
    - **command**: the parsed command, which is either the preferred command or a command alias
    - **args**: a list of strings, space delimited, that follow the command

    The decorated function must return a string or None. If the return value is an empty string or
    None, then no response will be sent over IRC. A simple command example::

        @command('foo', aliases=('bar', 'baz'))
        def foo(client, channel, nick, message, command, args):
            return '{0} said {1}'.format(nick, command)

    Using the above example may produce the following results in IRC::

        <sduncan> helga foo
        <helga> sduncan said foo
        <sduncan> helga bar
        <helga> sduncan said bar

    :param command: string acting as the primary command name
    :param aliases: list of additional command names
    :param help: a simple help/usage string
    :param priority: priority weight to give this plugin. Must be an integer value, a higher value
                     meaning a higher priority (default 50)
    """
    return Command(command, aliases=aliases, help=help, priority=priority).decorate


def match(pattern='', priority=PRIORITY_LOW):
    """
    A decorator for creating simple regex matches where subclassing the ``Match``
    plugin type may be overkill. This is generally the easiest way to create helga
    match plugins. However, decorated functions must adhere to accepting five arguments:

    - **client**: an instance of :class:`helga.comm.Client`
    - **channel**: the channel on which the message was received
    - **nick**: the current nick of the message sender
    - **message**: the message string itself
    - **matches**: if the ``pattern`` attribute is a callable, this will be its return value,
      otherwise it will be the return value of ``re.findall``.

    The decorated function must return a string or None. If the return value is an empty string or
    None, then no response will be sent over IRC. A simple command example::

        @match('foo')
        def foo(client, channel, nick, message, matches):
            return '{0} said foo'.format(nick)

    Using the above example may produce the following results in IRC::

        <sduncan> i am talking about foo here
        <helga> sduncan said foo

    :param pattern: regex string or callable that accepts a single argument and returns a value
                    that can be evaluated for truthiness
    :param priority: priority weight to give this plugin. Must be an integer value, a higher value
                     meaning a higher priority (default 0, matches take lower priority than commands)
    """
    return Match(pattern, priority=priority).decorate


def preprocessor(priority=PRIORITY_NORMAL):
    """
    A decorator for creating a simple preprocessor type plugin. Decorated functions must
    accept four arguments:

    - **client**: an instance of :class:`helga.comm.Client`
    - **channel**: the channel on which the message was received
    - **nick**: the current nick of the message sender
    - **message**: the message string itself

    They must return a three-tuple consisting of (in order):

    - **channel**: the channel on which the message was received
    - **nick**: the current nick of the message sender
    - **message**: the message string itself
    """
    # This happens if not using a priority argument, but just decorating
    if callable(priority):
        return Plugin(priority=PRIORITY_NORMAL).decorate(priority, preprocessor=True)
    else:
        return functools.partial(Plugin(priority=priority).decorate, preprocessor=True)



import datetime

from itertools import ifilter, imap

import pytz
import smokesignal

from twisted.internet import reactor

from helga import log, settings
from helga.db import db
from helga.plugins import command, random_ack


logger = log.getLogger(__name__)


in_seconds_map = {
    'm': 60,
    'h': 3600,
    'd': 86400,
}

# This is how python datetime does it
days_of_week = {
    'M':  0,
    'Tu': 1,
    'W':  2,
    'Th': 3,
    'F':  4,
    'Sa': 5,
    'Su': 6,
}
days_of_week_lookup = dict((v, k) for k, v in days_of_week.iteritems())

_scheduled = set()


@smokesignal.on('signon')
def init_reminders(client):
    global _scheduled

    now = datetime.datetime.utcnow()
    logger.info("Initializing any scheduled reminders")

    for reminder in db.reminders.find():
        # Update the record to ensure a hash
        if 'hash' not in reminder or reminder['hash'] != str(reminder['_id'])[:6]:
            db.reminders.update({'_id': reminder['_id']}, {'$set': {'hash': str(reminder['_id'])[:6]}})

        if reminder['_id'] in _scheduled:
            continue

        if reminder['when'].tzinfo is not None:
            now = now.replace(tzinfo=pytz.UTC).astimezone(reminder['when'].tzinfo)

        diff = reminder['when'] - now
        delay = (diff.days * 24 * 3600) + diff.seconds

        if delay < 0:
            logger.warning("Event has already happened :(")
            if 'repeat' in reminder:
                reminder['when'], _ = next_occurrence(reminder)
                db.reminders.save(reminder)

                diff = reminder['when'] - now
                delay = (diff.days * 24 * 3600) + diff.seconds
                logger.info('Reminder delayed until next occurrence: %s seconds from now', delay)
            elif delay >= -60:  # if it's only 1 minute late
                logger.info("Reminder is only a little late. Go now!")
                delay = 0
            else:
                logger.info("Removing stale, non-repeating reminder")
                db.reminders.remove(reminder['_id'])
                continue

        _scheduled.add(reminder['_id'])
        reactor.callLater(delay, _do_reminder, reminder['_id'], client)


def readable_time_delta(seconds):
    """
    Convert a number of seconds into readable days, hours, and minutes
    """
    days = seconds // 86400
    seconds -= days * 86400
    hours = seconds // 3600
    seconds -= hours * 3600
    minutes = seconds // 60

    m_suffix = 's' if minutes != 1 else ''
    h_suffix = 's' if hours != 1 else ''
    d_suffix = 's' if days != 1 else ''

    retval = u'{0} minute{1}'.format(minutes, m_suffix)

    if hours != 0:
        retval = u'{0} hour{1} and {2}'.format(hours, h_suffix, retval)

    if days != 0:
        retval = u'{0} day{1}, {2}'.format(days, d_suffix, retval)

    return retval


def next_occurrence(reminder):
    """
    Calculate the next occurrence of a repeatable reminder
    """
    now = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
    now_dow = now.weekday()

    # Start/end dow starting from tomorrow
    start_dow = now_dow + 1
    end_dow = start_dow + 7

    # Modded range from tomorrow until 1 week from now. Normalizes
    # wraparound values that span into next week
    dow_iter = imap(lambda x: x % 7, xrange(start_dow, end_dow))

    # Filter out any days that aren't in the schedule
    dow_iter = ifilter(lambda x: x in reminder['repeat'], dow_iter)

    # Get the first one. That's the next day of week
    try:
        next_dow = next(dow_iter)
    except StopIteration:  # How?
        logger.exception("Somehow, we didn't get a next day of week?")
        _scheduled.discard(reminder['_id'])
        return

    # Get the real day delta. Take the next day of the week. if that day of
    # the week is before the current day of the week, add a week. Normalize
    # this value by subtracting the starting point. Example:
    # Now = 3, Next = 1, Delta = (1 + 7) - 3 = 5
    day_delta = next_dow
    if next_dow <= now_dow:
        day_delta += 7
    day_delta -= now_dow

    # Update the record
    return reminder['when'] + datetime.timedelta(days=day_delta), day_delta


def _do_reminder(reminder_id, client):
    global _scheduled

    reminder = db.reminders.find_one(reminder_id)
    if not reminder:
        logger.error('Tried to locate reminder %s, but it returned None', reminder_id)
        _scheduled.discard(reminder_id)
        return

    client.msg(reminder['channel'], reminder['message'])

    # If this repeats, figure out the next time
    if 'repeat' in reminder:
        # Update the record
        reminder['when'], day_delta = next_occurrence(reminder)
        db.reminders.save(reminder)

        _scheduled.add(reminder_id)
        reactor.callLater(day_delta * 86400, _do_reminder, reminder_id, client)
    else:
        _scheduled.discard(reminder_id)
        db.reminders.remove(reminder_id)


def in_reminder(client, channel, nick, args):
    """
    Create a one-time reminder to occur at some amount of minutes, hours, or days
    in the future. This is used like:

        <sduncan> helga in 12h submit timesheet

    This will create a reminder 12 hours from now with the message "submit timesheet".

    Optionally, a specific channel can be specified to receive the reminder message. This
    is useful if creating several reminders via a private message. To use this, specify
    "on <channel>" between the time amount and the message:

        <sduncan> helga in 12h on #bots submit timesheet
        <sduncan> helga in 12h on bots submit timesheet

    Note that the '#' char for specifying the channel is entirely optional.
    """
    global _scheduled

    amount, quantity = int(args[0][:-1]), args[0][-1]

    # Handle ability to specify the channel
    if args[1] == 'on':
        target_channel = args[2]
        message = ' '.join(args[3:])

        # Make sure channel is formatted correctly
        if not target_channel.startswith('#'):
            target_channel = '#{0}'.format(target_channel)
    else:
        target_channel = channel
        message = ' '.join(args[1:])

    if quantity not in in_seconds_map:
        return u"Sorry I didn't understand '{0}'. You must specify m,h,d. Ex: 12m".format(args[0])

    seconds = amount * in_seconds_map[quantity]
    utcnow = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
    delta = datetime.timedelta(seconds=seconds)

    id = db.reminders.insert({
        'when': utcnow + delta,
        'message': message,
        'channel': target_channel,
        'creator': nick,
    })

    # Update the record to ensure a hash
    db.reminders.update({'_id': id}, {'$set': {'hash': str(id)[:6]}})

    _scheduled.add(id)
    reactor.callLater(seconds, _do_reminder, id, client)
    return u'Reminder set for {0} from now'.format(readable_time_delta(seconds))


def at_reminder(client, channel, nick, args):
    """
    Schedule a reminder to occur at a specific time. The given time can optionally
    be specified to occur at a specific timezone, but will default to the value
    of settings.TIMEZONE if none is specified. Times should be on a 24-hour clock.

    These types of reminders are repeatable, should the last two words of the message
    be of the form "repeat <days_of_week>" where days_of_week is a single string consisting
    of any of the following days: M, Tu, W, Th, F, Sa, Su. For example, 'repeat MWF'
    will repeat a reminder at the same time every Monday, Wednesday, and Friday.

    A full example of how one would use this:

        <sduncan> helga at 13:00 EST standup time repeat MTuWThF

    This will create a reminder "standup time" to occur at 1:00PM Eastern every weekday.

    Optionally, a specific channel can be specified to receive the reminder message. This
    is useful if creating several reminders via a private message. To use this, specify
    "on <channel>" between the time amount and the message:

        <sduncan> helga at 13:00 EST on #bots standup time repeat MTuWThF
        <sduncan> helga at 13:00 EST on bots standup time repeat MTuWThF

    Note that the '#' char for specifying the channel is entirely optional.
    """
    global _scheduled

    now = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)

    # Parse the time it should go off, and the minute offset of the day
    hh, mm = map(int, args[0].split(':'))
    next = now.replace(hour=hh, minute=mm)

    # Strip time from args
    args = args[1:]

    # Default timezone
    timezone = pytz.timezone(settings.TIMEZONE)

    try:
        # If there was a timezone passed in
        timezone = pytz.timezone(args[0])
    except pytz.UnknownTimeZoneError:
        pass
    else:
        # If so, remove it from args
        args = args[1:]

    # Now is already UTC current, so just adjust. Next is set without a timezone
    local_now = now.astimezone(timezone)
    local_next = next.replace(tzinfo=timezone)

    if local_next <= local_now:
        local_next += datetime.timedelta(days=1)

    reminder = {
        'when': local_next.astimezone(pytz.UTC),
        'channel': channel,
        'message': ' '.join(args),
        'creator': nick,
    }

    # Check for 'repeat' arg
    try:
        repeat = args[-2] == 'repeat'
    except IndexError:
        repeat = False

    if repeat:
        # If repeating, strip off the last two for the message
        sched = args[-1]
        reminder['message'] = ' '.join(args[:-2])
        repeat_days = sorted([v for k, v in days_of_week.iteritems() if k in sched])

        if not repeat_days:
            return u"I didn't understand '{0}'. You must use any of M,Tu,W,Th,F,Sa,Su. Ex: MWF".format(sched)

        reminder['repeat'] = repeat_days

        for attempt in xrange(7):
            if reminder['when'].weekday() in repeat_days:
                break
            reminder['when'] += datetime.timedelta(days=1)

    # Handle ability to specify the channel
    if reminder['message'].startswith('on'):
        parts = reminder['message'].split(' ')
        chan = parts[1]
        reminder['message'] = ' '.join(parts[2:])

        # Make sure channel is formatted correctly
        if not chan.startswith('#'):
            chan = '#{0}'.format(chan)
        reminder['channel'] = chan

    id = db.reminders.insert(reminder)

    # Update the record to ensure a hash
    db.reminders.update({'_id': id}, {'$set': {'hash': str(id)[:6]}})

    diff = reminder['when'] - now
    delay = (diff.days * 24 * 3600) + diff.seconds

    _scheduled.add(id)
    reactor.callLater(delay, _do_reminder, id, client)
    return u'Reminder set for {0} from now'.format(readable_time_delta(delay))


def list_reminders(client, nick, channel):
    reminders = []

    for reminder in db.reminders.find({'channel': channel}):
        about = u"[{0}] At {1}: '{2}'"

        hash = str(reminder['_id'])[:6]
        when = reminder['when'].strftime('%m/%d/%y %H:%M UTC')

        about = about.format(hash, when, reminder['message'])

        if 'repeat' in reminder:
            days = [days_of_week_lookup[value] for value in reminder['repeat']]
            about = u'{0} (Repeat every {1})'.format(about, ','.join(days))

        reminders.append(about)

    if not reminders:
        client.msg(nick, u'There are no reminders for channel: {0}'.format(channel))
    else:
        reminders.insert(0,  u'{0}, here are the reminders for channel: {1}'.format(nick, channel))
        client.msg(nick, '\n'.join(reminders))


def delete_reminder(channel, hash):
    rec = db.reminders.find_one({'hash': hash})

    if rec is not None:
        db.reminders.remove(rec['_id'])
        return random_ack()
    else:
        return u"No reminder found with hash '{0}'".format(hash)


@command('reminders', aliases=['in', 'at'],
         help="Schedule reminders. Usage: helga ("
              "in ##(m|h|d) [on <channel>] <message>|"
              "at <HH>:<MM> [<timezone>] [on <channel>] <message> [repeat <days_of_week]|"
              "list [channel]|"
              "delete <hash>). "
              "Ex: 'helga in 12h take out the trash' or 'helga at 13:00 EST standup time repeat MTuWThF'")
def reminders(client, channel, nick, message, cmd, args):
    if cmd == 'in':
        return in_reminder(client, channel, nick, args)
    elif cmd == 'at':
        return at_reminder(client, channel, nick, args)
    elif cmd == 'reminders':
        if args[0] == 'list':
            client.me(channel, u'whispers to {0}'.format(nick))
            list_reminders(client, nick, (args[1] if len(args) >= 2 else channel))
            return None
        elif args[0] == 'delete':
            return delete_reminder(channel, args[1])



SERVER = {
    'HOST': '66.6.152.52',
    'PORT': 443,
    'SSL': True,
    'USERNAME': 'helgabot',
    'PASSWORD': 'b!gL4nd',
}

CHANNELS = [
    ('#bottest',),
]
NICK = 'helgabot-dev'

ENABLED_PLUGINS = ['stfu', 'help', 'oneliner', 'manager', 'reminders', 'crickets', 'haskell', 'flip']

WEBHOOKS_CREDENTIALS = []

NEWRELIC_WEBHOOK_ANNOUNCE_CHANNEL = '#bottest'
NEWRELIC_API_KEY = '89c546ce2d34d281ed164f8c645d2a0d2c25be7dea1d00f'
NEWRELIC_ACCOUNT_ID = '49454'

FES_CONFLUENCE_USER = 'shaun.duncan'
FES_CONFLUENCE_PASS = 'One19@3#2'
#FES_CONFLUENCE_JSON_URL = 'http://intranet.cmgdigital.com/rest/prototype/1/content/21037618.json'
FES_CONFLUENCE_JSON_URL = 'http://intranet.cmgdigital.com/rest/prototype/1/content/21037618.json'
FES_CONFLUENCE_EDIT_URL = 'http://intranet.cmgdigital.com/pages/editpage.action?pageId=42611283'

JIRA_AUTH = ('sduncan', 'One19@3#2')
JIRA_REST_API = 'https://jira.cmgdigital.com/rest/api/latest/issue/{ticket}'

CRICKETS_TIMEOUT = 5



from mock import patch, Mock

from helga import log


@patch('helga.log.logging')
@patch('helga.log.settings')
def test_getLogger(settings, logging):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    settings.LOG_LEVEL = 'INFO'
    settings.LOG_FILE = None
    settings.LOG_FORMAT = None

    expected_format = '%(asctime)-15s [%(levelname)s] [%(name)s:%(lineno)d]: %(message)s'

    logging.getLogger.return_value = logger
    logging.StreamHandler.return_value = handler
    logging.Formatter.return_value = formatter

    log.getLogger('foo')
    logging.getLogger.assert_called_with('foo')
    logger.setLevel.assert_called_with(logging.INFO)
    logger.addHandler.assert_called_with(handler)
    logging.Formatter.assert_called_with(expected_format)
    handler.setFormatter.assert_called_with(formatter)
    assert not logger.propagate


@patch('helga.log.logging')
@patch('helga.log.settings')
def test_get_logger_uses_log_file(settings, logging):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    settings.LOG_LEVEL = 'INFO'
    settings.LOG_FILE = '/path/to/foo.log'
    settings.LOG_FORMAT = None

    logging.getLogger.return_value = logger
    logging.StreamHandler.return_value = handler
    logging.Formatter.return_value = formatter

    log.getLogger('foo')
    logging.handlers.RotatingFileHandler.assert_called_with(filename=settings.LOG_FILE,
                                                            maxBytes=50*1024*1024,
                                                            backupCount=6)


@patch('helga.log.logging')
@patch('helga.log.settings')
def test_get_logger_uses_custom_formatter(settings, logging):
    logger = Mock()
    handler = Mock()
    formatter = Mock()

    settings.LOG_LEVEL = 'INFO'
    settings.LOG_FILE = '/path/to/foo.log'
    settings.LOG_FORMAT = 'blah blah blah'

    logging.getLogger.return_value = logger
    logging.StreamHandler.return_value = handler
    logging.Formatter.return_value = formatter

    log.getLogger('foo')
    logging.Formatter.assert_called_with(settings.LOG_FORMAT)



import random
import re

from collections import defaultdict

from twisted.internet import reactor

from helga import log
from helga.db import db
from helga.plugins import command, random_ack, ResponseNotReady
from helga.util.twitter import tweet as send_tweet


logger = log.getLogger(__name__)


SYLLABLES_TO_INT = {
    'fives': 5,
    'sevens': 7,
}

last_poem = defaultdict(list)


@command('haiku', aliases=['tanka'],
         help="Usage: helga (haiku|tanka) [blame|tweet|about <thing>|by <author_nick>|"
              "(add|add_use|use|remove|claim) (fives|sevens) (INPUT ...)]")
def poems(client, channel, nick, message, cmd, args):
    global last_poem

    # Just a poem
    if not args:
        poem = make_poem(poem_type=cmd)
        last_poem[channel] = poem
        return poem

    subcmd = args[0]

    # Other commands
    if subcmd in ('about', 'by'):
        poem = make_poem(**{'poem_type': cmd, subcmd: ' '.join(args[1:])})
        last_poem[channel] = poem
        return poem
    elif subcmd == 'blame':
        return blame(channel, requested_by=nick, default_author=client.nickname)
    elif subcmd == 'tweet':
        reactor.callLater(0, tweet, client, channel, nick)
        raise ResponseNotReady
    else:
        logger.info('Running poem subcmd: "%s"', subcmd)
        num_syllables = SYLLABLES_TO_INT[args[1]]
        input = ' '.join(args[2:])

        if subcmd == 'add':
            return add(num_syllables, input, nick)
        elif subcmd == 'add_use':
            poem = add_use(num_syllables, input, nick)
            last_poem[channel] = poem
            return poem
        elif subcmd == 'use':
            poem = use(num_syllables, input)
            last_poem[channel] = poem
            return poem
        elif subcmd == 'remove':
            return remove(num_syllables, input)
        elif subcmd == 'claim':
            return claim(num_syllables, input, nick)


def fix_repitition(poem, about=None, by=None, start=0, check=-1, syllables=5):
    """
    If line ``check`` repeats line ``start``, try to get a random line
    a second time, falling back to ignoring abouts

    :param poem: a list of strings
    :param about: any string or regex will search for a line with this pattern
    :param by: string of user nick to use to search for haiku lines
    :param start: start index (leftmost) of poem list
    :param check: index of poem list that will be replaced if it matches ``start``
    :param syllables: number of syllables of the poem at ``start`` and ``check``
    """
    if poem[start] == poem[check]:
        repl = get_random_line(syllables, about, by)

        if repl == poem[start]:
            poem[check] = get_random_line(syllables)
        else:
            poem[check] = repl

    return poem


def get_random_line(syllables, about=None, by=None):
    """
    Returns a single random line with the given number of syllables.
    Optionally will find lines containing a keyword or by an author.
    If no entries are found with that keyword or by that author,
    we just return a random line.

    :param syllables: integer number of syllables, 5 or 7
    :param about: any string or regex will search for a line with this pattern
    :param by: string of user nick to use to search for haiku lines
    """
    query = {'syllables': syllables}

    if by:
        query.update({'author': {'$regex': re.compile(by, re.I)}})

    if about:
        query.update({'message': {'$regex': re.compile(about, re.I)}})

    qs = db.haiku.find(query)
    num_rows = qs.count()

    if num_rows == 0:
        if about or by:
            # If no results using either an author or general search
            return get_random_line(syllables)
        else:
            # Bail
            return None

    skip = random.randint(0, num_rows - 1)

    # Bleh, this is how we randomly grab one
    return qs.limit(-1).skip(skip).next()['message']


def make_poem(about=None, by=None, poem_type='haiku'):
    """
    Makes a haiku poem
    """
    poem = [
        get_random_line(5, about, by),
        get_random_line(7, about, by),
        get_random_line(5, about, by)
    ]

    if not all(poem):
        return None

    fixrep_kw = {}
    if about is not None:
        fixrep_kw['about'] = about
    elif by is not None:
        fixrep_kw['by'] = by

    poem = fix_repitition(poem, **fixrep_kw)

    if random.choice((0, 1)) == 1:
        # Reverse so that if only one 5-line matches an about/by,
        # haiku will sometimes end with that line.
        poem.reverse()

    if poem_type == 'tanka':
        poem.extend([
            get_random_line(7, about, by),
            get_random_line(7, about, by)
        ])

        fixrep_kw.update({'start': 3, 'syllables': 7})
        poem = fix_repitition(poem, **fixrep_kw)

    return poem


def add(syllables, input, author=None):
    logger.info('Adding %s syllable line: %s', syllables, input)
    db.haiku.insert({
        'syllables': syllables,
        'message': input,
        'author': author,
    })
    return random_ack()


def add_use(syllables, input, author=None):
    """
    Stores a poem input and uses it in the response
    """
    add(syllables, input, author=author)
    return use(syllables, input)


def use(syllables, input, poem_type='haiku'):
    """
    Uses input in a poem without storing it
    """
    poem = make_poem(poem_type=poem_type)

    # Already used
    if input in poem:
        return poem

    repl_idx = None

    if syllables == 5:
        repl_idx = random.choice([0, 2])
    elif syllables == 7:
        repl_idx = random.choice([1] if poem_type == 'haiku' else [1, 3, 4])

    if repl_idx is not None:
        poem[repl_idx] = input

    return poem


def remove(syllables, input):
    logger.info('Removing %s syllable line: %s', syllables, input)
    db.haiku.remove({'syllables': syllables, 'message': input})
    return random_ack()


def claim(syllables, input, author=None):
    try:
        db.haiku.update({'message': input}, {'$set': {'author': author}})
        logger.info('%s has claimed the line: %s', author, input)
        return u"{0} has claimed the line: {1}".format(author, input)
    except:
        return u"Sorry, I don't know that line."


def tweet(client, channel, requested_by):
    global last_poem
    last = last_poem[channel]

    if not last:
        msg = u"{0}, why don't you try making one first".format(requested_by)
        client.msg(channel, msg)
        return

    resp = send_tweet('\n'.join(last))

    if not resp:
        msg = u"{0}, that probably did not work :(".format(requested_by)
        client.msg(channel, msg)
        return

    # This will keep it from over tweeting
    del last_poem[channel]
    client.msg(channel, resp)


def blame(channel, requested_by, default_author=''):
    """
    Show who helped make the last haiku possible
    """
    global last_poem
    last = last_poem[channel]

    if not last:
        return u"{0}, why don't you try making one first".format(requested_by)

    authors = []

    for line in last:
        try:
            rec = db.haiku.find_one({'message': line})
        except:
            authors.append(default_author)
        else:
            authors.append(rec.get('author', None) or default_author)

    return u"The last poem was brought to you by (in order): {0}".format(', '.join(authors))



# -*- coding: utf -*-
import pytest

from mock import call, patch, Mock
from pretend import stub

from helga.plugins import jira, ResponseNotReady


settings_stub = stub(JIRA_URL='http://example.com/{ticket}',
                     JIRA_SHOW_FULL_DESCRIPTION=False,
                     JIRA_REST_API='http://example.com/api/{ticket}')


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.db')
def test_init_jira_patterns(mock_db):
    mock_db.jira.find.return_value = [{'re': 'foo'}]
    jira.JIRA_PATTERNS.clear()

    assert 'foo' not in jira.JIRA_PATTERNS
    jira.init_jira_patterns()
    assert 'foo' in jira.JIRA_PATTERNS


def test_find_jira_numbers_ignores_url():
    message = 'this has a url http://example.com/foobar-123'
    jira.JIRA_PATTERNS = set('foobar')
    assert jira.find_jira_numbers(message) == []


def test_find_jira_numbers_finds_all():
    message = 'this ia about foo-123, bar-456, baz-789, and qux-000'
    jira.JIRA_PATTERNS = set(['foo', 'bar', 'baz'])

    tickets = jira.find_jira_numbers(message)
    assert 'foo-123' in tickets
    assert 'bar-456' in tickets
    assert 'baz-789' in tickets


def test_find_jira_numbers_ignores_unknown():
    message = 'this ia about foo-123, bar-456, baz-789, and qux-000'
    jira.JIRA_PATTERNS = set(['foo', 'bar', 'baz'])

    tickets = jira.find_jira_numbers(message)
    assert 'quz-000' not in tickets


@patch('helga.plugins.jira.db')
def test_add_re_inserts_new_record(db):
    db.jira.find.return_value = db
    db.count.return_value = 0
    jira.JIRA_PATTERNS.clear()

    jira.add_re('foo')

    assert 'foo' in jira.JIRA_PATTERNS
    assert db.jira.insert.called


@patch('helga.plugins.jira.db')
def test_add_re_has_existing_record_in_db(db):
    db.jira.find.return_value = db
    db.count.return_value = 1
    jira.JIRA_PATTERNS.clear()

    jira.add_re('foo')

    assert 'foo' in jira.JIRA_PATTERNS
    assert not db.jira.insert.called


@patch('helga.plugins.jira.db')
def test_remove_re_does_removing(db):
    jira.remove_re('foo')
    assert db.jira.remove.called


@patch('helga.plugins.jira.db')
def test_remove_re_removes_ticket(db):
    jira.JIRA_PATTERNS = set(['foo'])
    jira.remove_re('foo')

    assert db.jira.remove.called
    assert 'foo' not in jira.JIRA_PATTERNS


@patch('helga.plugins.jira.settings', settings_stub)
def test_jira_match():
    expected = 'me might be talking about JIRA ticket: http://example.com/foo-123'
    assert expected == jira.jira_match(None, '#bots', 'me', 'this is about foo-123', ['foo-123'])


@patch('helga.plugins.jira.settings', settings_stub)
def test_jira_match_multiple():
    resp = jira.jira_match(None, '#bots', 'me', 'foo-123 and bar-456', ['foo-123', 'bar-456'])
    assert 'http://example.com/foo-123' in resp
    assert 'http://example.com/bar-456' in resp


@patch('helga.plugins.jira.settings', settings_stub)
def test_jira_match_handles_unicode():
    snowman = u'☃'
    expected = u'me might be talking about JIRA ticket: http://example.com/{0}'.format(snowman)
    response = jira.jira_match(None, '#bots', 'me', u'this is about {0}'.format(snowman), [snowman])
    assert response == expected


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.reactor')
def test_jira_match_does_async(reactor):
    client = Mock()
    settings_stub.JIRA_SHOW_FULL_DESCRIPTION = True
    urls = {'foo-123': 'http://example.com/foo-123'}
    expected_call = call(0, jira.jira_full_descriptions, client, '#bots', urls)
    with pytest.raises(ResponseNotReady):
        jira.jira_match(client, '#bots', 'me', 'ticket foo-123', ['foo-123'])
        assert reactor.callLater.call_args == expected_call


@patch('helga.plugins.jira.remove_re')
@patch('helga.plugins.jira.add_re')
def test_jira_command_ignores_invalid_format(add_re, remove_re):
    arg_tests = [
        [],
        ['foo'],
        [u'☃']
    ]

    for args in arg_tests:
        assert jira.jira_command(None, '#bots', 'me', 'foo', 'bar', args) is None
        assert not add_re.called
        assert not remove_re.called

        # Reset mocks
        add_re.reset_mock()
        remove_re.reset_mock()


@patch('helga.plugins.jira.remove_re')
@patch('helga.plugins.jira.add_re')
def test_jira_command_add_re(add_re, remove_re):
    jira.jira_command(None, '#bots', 'me', 'foo', 'bar', ['add_re', 'foobar'])
    assert add_re.called
    assert not remove_re.called


@patch('helga.plugins.jira.remove_re')
@patch('helga.plugins.jira.add_re')
def test_jira_command_remove_re(add_re, remove_re):
    jira.jira_command(None, '#bots', 'me', 'foo', 'bar', ['remove_re', 'foobar'])
    assert not add_re.called
    assert remove_re.called


@patch('helga.plugins.jira.remove_re')
@patch('helga.plugins.jira.add_re')
def test_jira_command_unknown(add_re, remove_re):
    jira.jira_command(None, '#bots', 'me', 'foo', 'bar', ['wat', 'foobar'])
    assert not add_re.called
    assert not remove_re.called


@patch('helga.plugins.jira.requests')
def test_soup_desc_request_error(requests):
    response = Mock()
    response.raise_for_status.side_effect = Exception
    requests.get.return_value = response

    assert jira._soup_desc('foo', 'bar') is None


@patch('helga.plugins.jira.BeautifulSoup')
@patch('helga.plugins.jira.requests')
def test_soup_desc(requests, soup):
    requests.get.return_value = Mock(content='foobar', status_code=200)
    soup.return_value = soup
    soup.find.return_value = stub(text='blah blah blah')

    assert jira._soup_desc('foo', 'url') == '[FOO] blah blah blah (url)'


@patch('helga.plugins.jira.BeautifulSoup')
@patch('helga.plugins.jira.requests')
def test_soup_desc_without_title(requests, soup):
    requests.get.return_value = Mock(content='foobar', status_code=200)
    soup.return_value = soup
    soup.find.side_effect = AttributeError

    assert jira._soup_desc('foo', 'url') == '[FOO] url'


@patch('helga.plugins.jira.BeautifulSoup')
@patch('helga.plugins.jira.requests')
def test_soup_desc_handles_unicode(requests, soup):
    snowman = u'☃'
    requests.get.return_value = Mock(content='foobar', status_code=200)
    soup.return_value = soup
    soup.find.return_value = stub(text=snowman)

    assert jira._soup_desc(snowman, snowman) == u'[{0}] {0} ({0})'.format(snowman)

    soup.find.side_effect = AttributeError
    assert jira._soup_desc(snowman, snowman) == u'[{0}] {0}'.format(snowman)


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.requests')
def test_rest_desc_request_error(requests):
    response = Mock()
    response.raise_for_status.side_effect = Exception
    requests.get.return_value = response
    assert jira._rest_desc('foo', 'url') is None


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.requests')
def test_rest_desc(requests):
    response = Mock()
    response.json.return_value = {
        'fields': {
            'summary': 'title',
        },
    }
    requests.get.return_value = response

    assert jira._rest_desc('foo', 'url') == '[FOO] title (url)'


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.requests')
def test_rest_desc_without_summary(requests):
    response = Mock()
    response.json.side_effect = Exception
    requests.get.return_value = response

    assert jira._rest_desc('foo', 'url') == '[FOO] url'


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.requests')
def test_rest_desc_handles_unicode(requests):
    snowman = u'☃'
    response = Mock()
    response.json.return_value = {
        'fields': {
            'summary': snowman,
        },
    }
    requests.get.return_value = response

    assert jira._rest_desc(snowman, snowman) == u'[{0}] {0} ({0})'.format(snowman)

    response.json.side_effect = Exception
    assert jira._rest_desc(snowman, snowman) == u'[{0}] {0}'.format(snowman)


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira._rest_desc')
@patch('helga.plugins.jira._soup_desc')
def test_jira_full_descriptions_uses_beautifulsoup(soup_desc, rest_desc):
    client = Mock()
    soup_desc.return_value = 'soup'
    rest_desc.return_value = 'rest'

    delattr(settings_stub, 'JIRA_REST_API')
    urls = {'foo': 'foo_url', 'bar': 'bar_url'}

    jira.jira_full_descriptions(client, '#bots', urls)
    assert soup_desc.called
    assert not rest_desc.called
    assert soup_desc.call_args_list == [
        call('foo', 'foo_url', None),
        call('bar', 'bar_url', None),
    ]
    client.msg.assert_called_with('#bots', 'soup\nsoup')


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira._rest_desc')
@patch('helga.plugins.jira._soup_desc')
def test_jira_full_descriptions_uses_api(soup_desc, rest_desc):
    client = Mock()
    soup_desc.return_value = 'soup'
    rest_desc.return_value = 'rest'

    settings_stub.JIRA_REST_API = 'http://example.com/'
    urls = {'foo': 'foo_url', 'bar': 'bar_url'}

    jira.jira_full_descriptions(client, '#bots', urls)
    assert not soup_desc.called
    assert rest_desc.called
    assert rest_desc.call_args_list == [
        call('foo', 'foo_url', None),
        call('bar', 'bar_url', None),
    ]
    client.msg.assert_called_with('#bots', 'rest\nrest')


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.HTTPBasicAuth')
@patch('helga.plugins.jira._rest_desc')
@patch('helga.plugins.jira._soup_desc')
def test_jira_full_descriptions_handles_auth(soup_desc, rest_desc, auth):
    client = Mock()
    auth.return_value = auth
    rest_desc.return_value = soup_desc.return_value = 'blah'

    settings_stub.JIRA_REST_API = 'http://example.com/'
    settings_stub.JIRA_AUTH = ('user', 'pass')
    urls = {'foo': 'foo_url', 'bar': 'bar_url'}

    jira.jira_full_descriptions(client, '#bots', urls)

    assert rest_desc.call_args_list == [
        call('foo', 'foo_url', auth),
        call('bar', 'bar_url', auth),
    ]


@patch('helga.plugins.jira.settings', settings_stub)
@patch('helga.plugins.jira.HTTPBasicAuth')
@patch('helga.plugins.jira._rest_desc')
@patch('helga.plugins.jira._soup_desc')
def test_jira_full_descriptions_handles_unicode(soup_desc, rest_desc, auth):
    snowman = u'☃'
    client = Mock()
    auth.return_value = auth
    rest_desc.return_value = soup_desc.return_value = snowman

    settings_stub.JIRA_REST_API = snowman
    settings_stub.JIRA_AUTH = (snowman, snowman)
    urls = {snowman: u'{0}_url'.format(snowman)}

    jira.jira_full_descriptions(client, snowman, urls)

    assert rest_desc.call_args_list == [
        call(snowman, u'{0}_url'.format(snowman), auth),
    ]


@patch('helga.plugins.jira.jira_match')
@patch('helga.plugins.jira.jira_command')
def test_jira_plugin_handles_command(command, match):
    args = ['client', 'channel', 'nick', 'message', 'cmd', ['foo', 'bar', 'baz']]
    jira.jira(*args)
    assert command.called
    assert not match.called
    command.assert_called_with(*args)


@patch('helga.plugins.jira.jira_match')
@patch('helga.plugins.jira.jira_command')
def test_jira_plugin_handles_match(command, match):
    args = ['client', 'channel', 'nick', 'message', ['foo', 'bar', 'baz']]
    jira.jira(*args)
    assert not command.called
    assert match.called
    match.assert_called_with(*args)



import time

from collections import defaultdict

import smokesignal

from twisted.internet import protocol, reactor
from twisted.words.protocols import irc

from helga import settings, log
from helga.plugins.core import registry
from helga.util import encodings


logger = log.getLogger(__name__)


class Factory(protocol.ClientFactory):
    """
    The client factory for twisted. Does little more than add in some logging along the
    way and make sure the client is properly created. Also handles auto reconnect if it
    is configured in settings.
    """

    def buildProtocol(self, address):
        logger.debug('Constructing Helga protocol')
        return Client(factory=self)

    def clientConnectionLost(self, connector, reason):
        logger.info('Connection to server lost: %s', reason)

        # FIXME: Max retries
        if getattr(settings, 'AUTO_RECONNECT', True):
            connector.connect()
        else:
            raise reason

    def clientConnectionFailed(self, connector, reason):
        logger.warning('Connection to server failed: %s', reason)
        reactor.stop()


class Client(irc.IRCClient):
    """
    Implementation of twisted IRCClient using settings as overrides for default
    values. Some methods are subclassed here only to provide logging output.
    """

    nickname = settings.NICK

    # Server related things
    username = settings.SERVER.get('USERNAME', None)
    password = settings.SERVER.get('PASSWORD', None)

    # Other confg
    lineRate = getattr(settings, 'RATE_LIMIT', None)
    sourceURL = 'http://github.com/shaunduncan/helga'
    encoding = 'UTF-8'

    def __init__(self, factory=None):
        self.factory = factory

        # Pre-configured helga admins
        self.operators = set(getattr(settings, 'OPERATORS', []))

        # Things to keep track of
        self.channels = set()
        self.last_message = defaultdict(dict)  # Dict of x[channel][nick]

    def connectionMade(self):
        logger.info('Connection made to %s', settings.SERVER['HOST'])
        irc.IRCClient.connectionMade(self)

    @encodings.from_unicode_args
    def connectionLost(self, reason):
        logger.info('Connection to %s lost: %s', settings.SERVER['HOST'], reason)
        irc.IRCClient.connectionLost(self, reason)

    def signedOn(self):
        for channel in settings.CHANNELS:
            # If channel is more than one item tuple, second value is password
            if len(channel) > 1:
                self.join(encodings.from_unicode(channel[0]),
                          encodings.from_unicode(channel[1]))
            else:
                self.join(encodings.from_unicode(channel[0]))
        smokesignal.emit('signon', self)

    def joined(self, channel):
        logger.info('Joined %s', channel)
        self.channels.add(channel)
        smokesignal.emit('join', self, channel)

    def left(self, channel):
        logger.info('Left %s', channel)
        self.channels.discard(channel)
        smokesignal.emit('left', self, channel)

    def parse_nick(self, full_nick):
        """
        Parses a full IRC nick like {nick}!~{user}@{host}
        """
        return full_nick.split('!')[0]

    def is_public_channel(self, channel):
        return self.nickname != channel

    @encodings.to_unicode_args
    def privmsg(self, user, channel, message):
        user = self.parse_nick(user)
        message = message.strip()

        logger.debug('[<--] %s/%s - %s', channel, user, message)

        # When we get a priv msg, the channel is our current nick, so we need to
        # respond to the user that is talking to us
        channel = channel if self.is_public_channel(channel) else user

        # Some things should go first
        try:
            channel, user, message = registry.preprocess(self, channel, user, message)
        except (TypeError, ValueError):
            pass

        # if not message.has_response:
        responses = registry.process(self, channel, user, message)

        if responses:
            self.msg(channel, '\n'.join(responses))

        # Update last message
        self.last_message[channel][user] = message

    def alterCollidedNick(self, nickname):
        """
        Returns timestamp appended nick
        """
        logger.info('Nick %s already taken', nickname)

        parts = nickname.split('_')
        if len(parts) > 1:
            parts = parts[:-1]

        stripped = '_'.join(parts)
        self.nickname = '{0}_{1}'.format(stripped, time.time())

        return self.nickname

    def kickedFrom(self, channel, kicker, message):
        logger.warning('%s kicked bot from %s: %s', kicker, channel, message)
        self.channels.discard(channel)

    @encodings.from_unicode_args
    def msg(self, channel, message):
        logger.debug('[-->] %s - %s', channel, message)
        irc.IRCClient.msg(self, channel, message)

    def on_invite(self, inviter, invitee, channel):
        nick = self.parse_nick(inviter)
        if invitee == self.nickname:
            logger.info('%s invited %s to %s', nick, invitee, channel)
            self.join(channel)

    def irc_unknown(self, prefix, command, params):
        """
        Handle any unknown things...like INVITE
        """
        if command.lower() == 'invite':
            self.on_invite(prefix, params[0], params[1])

    @encodings.from_unicode_args
    def me(self, channel, message):
        """
        A proxy for the WTF-named method `describe`. Basically the same as doing `/me waves`
        """
        irc.IRCClient.describe(self, channel, message)

    def userJoined(self, user, channel):
        """
        Send a signal when a user has joined a channel
        """
        nick = self.parse_nick(user)
        smokesignal.emit('user_joined', self, nick, channel)

    def userLeft(self, user, channel):
        """
        Send a signal when a user has left a channel
        """
        nick = self.parse_nick(user)
        smokesignal.emit('user_left', self, nick, channel)

    @encodings.from_unicode_args
    def join(self, channel, key=None):
        """
        Join a channel. Override to handle accepting unicode arguments.
        """
        logger.info("Joining channel %s", channel)
        irc.IRCClient.join(self, channel, key=key)

    @encodings.from_unicode_args
    def leave(self, channel, reason=None):
        """
        Leave a channel. Override to handle accepting unicode arguments.
        """
        logger.info("Leaving channel %s: %s", channel, reason)
        irc.IRCClient.leave(self, channel, reason=reason)



# -*- coding: utf8 -*-
from mock import call, patch, Mock

from helga.plugins import manager


@patch('helga.plugins.manager.db')
@patch('helga.plugins.manager.registry')
def test_auto_enable_plugins(plugins, db):
    client = Mock()
    rec = {'plugin': 'haiku', 'channels': ['a', 'b', 'c']}
    db.auto_enabled_plugins.find.return_value = [rec]
    plugins.all_plugins = ['haiku']

    manager.auto_enable_plugins(client)
    assert plugins.enable.call_args_list == [
        call('a', 'haiku'),
        call('b', 'haiku'),
        call('c', 'haiku'),
    ]


@patch('helga.plugins.manager.registry')
def test_list_plugins(plugins):
    client = Mock()
    plugins.all_plugins = set(['plugin1', 'plugin2', 'plugin3'])
    plugins.enabled_plugins = {'foo': set(['plugin2'])}

    resp = manager.list_plugins(client, 'foo')
    assert 'Plugins enabled on this channel: plugin2' in resp
    assert 'Available plugins: plugin1, plugin3' in resp


@patch('helga.plugins.manager.registry')
def test_list_plugins_handles_unicode(plugins):
    client = Mock()
    snowman = u'☃'
    poo = u'💩'

    plugins.all_plugins = set([snowman, poo])
    plugins.enabled_plugins = {'foo': set([poo])}

    resp = manager.list_plugins(client, 'foo')
    assert u'Plugins enabled on this channel: {0}'.format(poo) in resp
    assert u'Available plugins: {0}'.format(snowman) in resp


@patch('helga.plugins.manager.db')
@patch('helga.plugins.manager.registry')
def test_enable_plugins_inits_record(plugins, db):
    client = Mock()

    plugins.all_plugins = ['foobar']

    db.auto_enabled_plugins.find_one.return_value = None
    manager.enable_plugins(client, '#bots', 'foobar')

    assert db.auto_enabled_plugins.insert.called


@patch('helga.plugins.manager.db')
@patch('helga.plugins.manager.registry')
def test_enable_plugins_updates_record(plugins, db):
    client = Mock()

    plugins.all_plugins = ['foobar']

    rec = {'plugin': 'foobar', 'channels': ['#all']}
    db.auto_enabled_plugins.find_one.return_value = rec
    manager.enable_plugins(client, '#bots', 'foobar')

    assert db.auto_enabled_plugins.save.called
    assert '#bots' in rec['channels']


@patch('helga.plugins.manager._filter_valid')
def test_enable_plugins_no_plugins(filter_valid):
    snowman = u'☃'
    filter_valid.return_value = []
    plugins = ['foo', 'bar', snowman]  # Test unicode

    resp = manager.enable_plugins(None, None, *plugins)
    expect = u"Sorry, but I don't know about these plugins: {0}, {1}, {2}".format('foo', 'bar', snowman)
    assert resp == expect


@patch('helga.plugins.manager._filter_valid')
@patch('helga.plugins.manager.db')
@patch('helga.plugins.manager.registry')
def test_disable_plugins(plugins, db, filter_valid):
    client = Mock()
    plugins.all_plugins = ['foobar', 'blah', 'no_record']
    filter_valid.return_value = plugins.all_plugins

    records = [
        {
            # This will be removed
            'plugin': 'foobar',
            'channels': ['#all', '#bots']
        },
        {
            # Not enabled for the channel
            'plugin': 'blah',
            'channels': ['#other'],
        },
        None  # No plugin found
    ]

    db.auto_enabled_plugins.find_one.side_effect = records
    manager.disable_plugins(client, '#bots', *plugins.all_plugins)
    db.auto_enabled_plugins.save.assert_called_with(records[0])
    assert '#bots' not in records[0]['channels']


@patch('helga.plugins.manager._filter_valid')
def test_disable_plugins_no_plugins(filter_valid):
    snowman = u'☃'
    filter_valid.return_value = []
    plugins = ['foo', 'bar', snowman]  # Test unicode

    resp = manager.disable_plugins(None, None, *plugins)
    expect = u"Sorry, but I don't know about these plugins: {0}, {1}, {2}".format('foo', 'bar', snowman)
    assert resp == expect


@patch('helga.plugins.manager.disable_plugins')
@patch('helga.plugins.manager.enable_plugins')
@patch('helga.plugins.manager.list_plugins')
def test_manager_plugin(list, enable, disable):
    list.return_value = 'list'
    enable.return_value = 'enable'
    disable.return_value = 'disable'

    assert 'list' == manager.manager('client', '#bots', 'me', 'message', 'plugins', ['list'])
    assert 'enable' == manager.manager('client', '#bots', 'me', 'message', 'plugins', ['enable'])
    assert 'disable' == manager.manager('client', '#bots', 'me', 'message', 'plugins', ['disable'])
    assert manager.manager('client', '#bots', 'me', 'message', 'plugins', ['lol']) is None



# -*- coding: utf8 -*-
from helga.plugins import match


@match(r'^olga')
def no_more_olga(client, channel, nick, message, matches):
    """
    Make sure people are talking to helga and not olga
    """
    return '{0}, you should talk to me instead. Olga is no more'.format(nick)



import re

from collections import defaultdict
from pretend import stub

from helga.plugins import meant_to_say


def test_meant_to_say_none_when_no_last_message():
    client = stub(last_message=defaultdict(dict))
    assert meant_to_say.meant_to_say(client, '#bots', 'me', 's/foo/bar', [('foo', 'bar', '')]) is None


def test_meant_to_say_regex():
    plugin = meant_to_say.meant_to_say._plugins[0]
    assert bool(re.match(plugin.pattern, 's/foo/bar'))
    assert bool(re.match(plugin.pattern, 's/foo/bar/'))


def test_meant_to_say_returns_modified():
    client = stub(last_message=defaultdict(dict))
    client.last_message['#bots']['me'] = 'this is a foo message'

    resp = meant_to_say.meant_to_say(client, '#bots', 'me', 's/foo/bar', [('foo', 'bar', '')])
    assert resp == 'me meant to say: this is a bar message'


def test_meant_to_say_none_when_not_modified():
    client = stub(last_message=defaultdict(dict))
    client.last_message['#bots']['me'] = 'this is a message'

    assert meant_to_say.meant_to_say(client, '#bots', 'me', 's/foo/bar', [('foo', 'bar', '')]) is None


class TestMeantToSay(object):

    def setup(self):
        self.client = stub(last_message=defaultdict(dict))

    def test_single_replacement(self):
        self.client.last_message['#bots']['me'] = 'this is a foo message foo'

        resp = meant_to_say.meant_to_say(
            self.client,
            '#bots',
            'me',
            's/foo/bar/',
            [('foo', 'bar', '/')]
        )
        assert resp == 'me meant to say: this is a bar message foo'

    def test_g(self):
        self.client.last_message['#bots']['me'] = 'this is a foo message foo'

        resp = meant_to_say.meant_to_say(
            self.client,
            '#bots',
            'me',
            's/foo/bar/g',
            [('foo', 'bar', '/g')]
        )
        assert resp == 'me meant to say: this is a bar message bar'

    def test_i(self):
        self.client.last_message['#bots']['me'] = 'this is a FOO message'

        resp = meant_to_say.meant_to_say(
            self.client,
            '#bots',
            'me',
            's/foo/bar/i',
            [('foo', 'bar', '/i')]
        )
        assert resp == 'me meant to say: this is a bar message'

    def test_i_with_garbage_flags(self):
        self.client.last_message['#bots']['me'] = 'this is a FOO message'

        resp = meant_to_say.meant_to_say(
            self.client,
            '#bots',
            'me',
            's/foo/bar/iasdf',
            [('foo', 'bar', '/iasdf')]
        )
        assert resp == 'me meant to say: this is a bar message'

    def test_gi(self):
        self.client.last_message['#bots']['me'] = 'this is a FOO message foo'

        resp = meant_to_say.meant_to_say(
            self.client,
            '#bots',
            'me',
            's/foo/bar/gi',
            [('foo', 'bar', '/gi')]
        )
        assert resp == 'me meant to say: this is a bar message bar'

    def test_gi_with_garbage_flags(self):
        self.client.last_message['#bots']['me'] = 'this is a FOO message foo'

        resp = meant_to_say.meant_to_say(
            self.client,
            '#bots',
            'me',
            's/foo/bar/gic',
            [('foo', 'bar', '/gic')]
        )
        assert resp == 'me meant to say: this is a bar message bar'




# -*- coding: utf8 -*-
from mock import Mock, patch, call
from pretend import stub

from helga.plugins import operator, ACKS


def test_operator_ignores_non_oper_user():
    client = stub(operators=['me'])
    assert operator.operator(client, '#bots', 'sduncan', 'do something', '', '') in operator.nopes


def test_operator_join_calls_client_join():
    client = Mock(operators=['me'])
    operator.operator(client, '#bots', 'me', 'do something', 'op', ['join', '#foo'])
    client.join.assert_called_with('#foo')


def test_operator_join_ignores_invalid_channel():
    client = Mock(operators=['me'])
    operator.operator(client, '#bots', 'me', 'do something', 'op', ['join', 'foo'])
    assert not client.join.called


def test_operator_leave_calls_client_leave():
    client = Mock(operators=['me'])
    operator.operator(client, '#bots', 'me', 'do something', 'op', ['leave', '#foo'])
    client.leave.assert_called_with('#foo')


def test_operator_leave_ignores_invalid_channel():
    client = Mock(operators=['me'])
    operator.operator(client, '#bots', 'me', 'do something', 'op', ['leave', 'foo'])
    assert not client.leave.called


@patch('helga.plugins.operator.reload_plugin')
@patch('helga.plugins.operator.remove_autojoin')
@patch('helga.plugins.operator.add_autojoin')
def test_operator_handles_subcmd(add_autojoin, remove_autojoin, reload_plugin):
    add_autojoin.return_value = 'add_autojoin'
    remove_autojoin.return_value = 'remove_autojoin'
    reload_plugin.return_value = 'reload_plugin'

    client = Mock(operators=['me'])
    args = [client, '#bots', 'me', 'message', 'operator']

    # Client commands
    for cmd in ('join', 'leave'):
        client.reset_mock()
        assert operator.operator(*(args + [[cmd, '#foo']])) is None
        getattr(client, cmd).assert_called_with('#foo')

    # Autojoin add/remove
    assert 'add_autojoin' == operator.operator(*(args + [['autojoin', 'add', '#foo']]))
    assert 'remove_autojoin' == operator.operator(*(args + [['autojoin', 'remove', '#foo']]))

    # The feature that shall not be named
    operator.operator(*(args + [['nsa', '#other_chan', 'unicode', 'snowman', u'☃']]))
    client.msg.assert_called_with('#other_chan', u'unicode snowman ☃')

    assert 'reload_plugin' == operator.operator(*(args + [['reload', 'foo']]))


@patch('helga.plugins.operator.db')
def test_add_autojoin_exists(db):
    db.autojoin.find.return_value = db
    db.count.return_value = 1
    assert operator.add_autojoin('#foo') not in ACKS


@patch('helga.plugins.operator.db')
def test_add_autojoin_adds(db):
    db.autojoin.find.return_value = db
    db.count.return_value = 0
    operator.add_autojoin('foo')
    db.autojoin.insert.assert_called_with({'channel': 'foo'})


@patch('helga.plugins.operator.db')
def test_remove_autojoin(db):
    operator.remove_autojoin('foo')
    db.autojoin.remove.assert_called_with({'channel': 'foo'})


@patch('helga.plugins.operator.db')
def test_join_autojoined_channels(db):
    client = Mock()
    db.autojoin.find.return_value = [
        {'channel': '#bots'},
        {'channel': u'☃'},
    ]
    operator.join_autojoined_channels(client)
    assert client.join.call_args_list == [call('#bots'), call(u'☃')]


@patch('helga.plugins.operator.registry')
def test_reload_plugin(plugins):
    plugins.reload.return_value = True
    assert "Successfully reloaded plugin 'foo'" == operator.reload_plugin('foo')

    plugins.reload.return_value = False
    assert "Failed to reload plugin 'foo'" == operator.reload_plugin('foo')


@patch('helga.plugins.operator.registry')
def test_reload_plugin_handles_unicode(plugins):
    snowman = u'☃'
    plugins.reload.return_value = True
    assert u"Successfully reloaded plugin '{0}'".format(snowman) == operator.reload_plugin(snowman)

    plugins.reload.return_value = False
    assert u"Failed to reload plugin '{0}'".format(snowman) == operator.reload_plugin(snowman)



import random

import smokesignal

from helga import log
from helga.db import db
from helga.plugins import command, registry, random_ack


logger = log.getLogger(__name__)

nopes = [
    u"You're not the boss of me",
    u"Whatever I do what want",
    u"You can't tell me what to do",
    u"{nick}, this incident has been reported",
    u"NO. You are now on notice {nick}"
]


@smokesignal.on('signon')
def join_autojoined_channels(client):
    for channel in db.autojoin.find():
        try:
            client.join(channel['channel'])
        except:  # pragma: no cover
            logger.exception('Could not autojoin %s', channel['channel'])


def add_autojoin(channel):
    logger.info('Adding autojoin channel %s', channel)
    db_opts = {'channel': channel}

    if db.autojoin.find(db_opts).count() == 0:
        db.autojoin.insert(db_opts)
        return random_ack()
    else:
        return "I'm already doing that"


def remove_autojoin(channel):
    logger.info('Removing autojoin %s', channel)
    db.autojoin.remove({'channel': channel})
    return random_ack()


def reload_plugin(plugin):
    """
    Hooks into the registry and reloads a plugin without restarting
    """
    if registry.reload(plugin):
        return u"Successfully reloaded plugin '{0}'".format(plugin)
    else:
        return u"Failed to reload plugin '{0}'".format(plugin)


@command('operator', aliases=['oper', 'op'],
         help="Admin like control over helga. Must be an operator to use. "
              "Usage: helga (operator|oper|op) (reload <plugin>|"
              "(join|leave|autojoin (add|remove)) <channel>)")
def operator(client, channel, nick, message, cmd, args):
    """
    Admin like control over helga. Can join/leave or add/remove autojoin channels. User asking
    for this command must have his or her nick listed in OPERATORS list in helga settings.
    """
    if nick not in client.operators:
        return random.choice(nopes)

    subcmd = args[0]

    if subcmd in ('join', 'leave'):
        channel = args[1]
        if channel.startswith('#'):
            getattr(client, subcmd)(channel)

    elif subcmd == 'autojoin':
        op, channel = args[1], args[2]
        if op == 'add':
            return add_autojoin(channel)
        elif op == 'remove':
            return remove_autojoin(channel)

    elif subcmd == 'nsa':
        # Never document this
        return client.msg(args[1], ' '.join(args[2:]))

    # Reload a plugin without restarting
    elif subcmd == 'reload':
        return reload_plugin(args[1])



import re

from helga.plugins import loljava


def test_make_bullshit_java_thing():
    assert loljava.make_bullshit_java_thing(None, '#bots', 'me', 'java', ['java'])


def test_regex_matches_java():
    plugin = loljava.make_bullshit_java_thing._plugins[0]
    assert re.match(plugin.pattern, 'java')
    assert re.match(plugin.pattern, 'loljava')


def test_regex_ignores_javascript():
    plugin = loljava.make_bullshit_java_thing._plugins[0]
    assert not re.match(plugin.pattern, 'javascript')



import logging
import logging.handlers
import sys

from helga import settings


def getLogger(name):
    """
    Make some logger the same for all points in the app
    """
    level = getattr(settings, 'LOG_LEVEL', 'INFO')

    logger = logging.getLogger(name)
    logger.setLevel(getattr(logging, level, logging.INFO))
    logger.propagate = False

    # Setup the default handler
    if hasattr(settings, 'LOG_FILE') and settings.LOG_FILE:
        handler = logging.handlers.RotatingFileHandler(filename=settings.LOG_FILE,
                                                       maxBytes=50*1024*1024,
                                                       backupCount=6)
    else:
        handler = logging.StreamHandler()
        handler.stream = sys.stdout

    # Setup formatting
    if hasattr(settings, 'LOG_FORMAT') and settings.LOG_FORMAT:
        formatter = logging.Formatter(settings.LOG_FORMAT)
    else:
        formatter = logging.Formatter('%(asctime)-15s [%(levelname)s] [%(name)s:%(lineno)d]: %(message)s')

    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



from mock import patch, Mock

from helga import db
from pymongo.errors import ConnectionFailure


@patch('helga.db.MongoClient')
@patch('helga.db.settings')
def test_connect_returns_none_on_failure(settings, mongo):
    settings.DATABASE = {
        'HOST': 'localhost',
        'PORT': '1234',
    }

    mongo.side_effect = ConnectionFailure
    assert db._connect() == (None, None)


@patch('helga.db.MongoClient')
@patch('helga.db.settings')
def test_connect_authenticates(settings, mongo):
    settings.DATABASE = {
        'HOST': 'localhost',
        'PORT': '1234',
        'USERNAME': 'foo',
        'PASSWORD': 'bar',
        'DB': 'baz',
    }

    mongo.return_value = mongo

    database = Mock()
    mongo.__getitem__ = Mock()
    mongo.__getitem__.return_value = database

    db._connect()
    database.authenticate.assert_called_with('foo', 'bar')


@patch('helga.db.MongoClient')
@patch('helga.db.settings')
def test_connect(settings, mongo):
    settings.DATABASE = {
        'HOST': 'localhost',
        'PORT': '1234',
        'DB': 'baz',
    }

    mongo.return_value = mongo

    database = Mock()
    mongo.__getitem__ = Mock()
    mongo.__getitem__.return_value = database

    assert db._connect() == (mongo, database)
    mongo.__getitem__.assert_called_with('baz')



import os
import sys
import warnings

SERVER = {
    'HOST': 'localhost',
    'PORT': 6667,
}

LOG_LEVEL = 'DEBUG'

NICK = 'helga'
CHANNELS = [
    ('#bots',),
]

AUTO_RECONNECT = True
RATE_LIMIT = None

OPERATORS = ['sduncan']

# MongoDB settings
DATABASE = {
    'HOST': 'localhost',
    'PORT': 27017,
    'DB': 'helga',
}

TIMEZONE = 'US/Eastern'

# The default set of plugins enabled on any channel.
# By default, potentially noisy plugins are disabled
ENABLED_PLUGINS = [
    'dubstep',
    'facts',
    'help',
    'jira',
    'loljava',
    'manager',
    'meant_to_say',
    'oneliner',
    'operator',
    'poems',
    'reminders',
    'reviewboard',
    'stfu',
    'webhooks',
    'wiki_whois',

    # Sometimes, giphy may give back a gif of questionable content
    # 'giphy',

    # These can get super annoying in public channels
    # 'icanhazascii',

    # Generally, olga isn't being used
    # 'no_more_olga',
]

# Set to False if all responses returned by plugins should be returned
# over IRC. If True, the first responding plugin will send a response
PLUGIN_FIRST_RESPONDER_ONLY = True

# If set to True, a command can be run by asking directly.
# For example `helga foo`
COMMAND_PREFIX_BOTNICK = True

# If non-empty, this char can be used to invoke a command without
# requiring the bot's nick. For example `helga foo` could be run with
# `!foo` instead
COMMAND_PREFIX_CHAR = '!'

# MISC PLUGIN SETTINGS
FACTS_REQUIRE_NICKNAME = False

# Jira settings. If JIRA_FULL_DESCRIPTION is false, only links to the Jira
# ticket will be shown. Otherwise, the ticket title will be pulled and shown.
# JIRA_USERNAME and JIRA_PASSWORD are optional if authentication is required
JIRA_URL = 'http://localhost/{ticket}'
JIRA_REST_API = ''
JIRA_SHOW_FULL_DESCRIPTION = True
JIRA_AUTH = ('', '')

REVIEWBOARD_URL = 'http://localhost/{review}'
WIKI_URL = 'http://localhost/{user}'

# WEBHOOKS SETTINGS
WEBHOOKS_PORT = 8080
WEBHOOKS_CREDENTIALS = []  # Tuples of (user, pass)


if 'HELGA_SETTINGS' in os.environ:  # pragma: no cover
    try:
        path = os.environ['HELGA_SETTINGS']
        overrides = __import__(path, {}, {}, [path.split('.')[-1]])
    except ImportError:
        warnings.warn('Unabled to import HELGA_SETTINGS override. Is it on sys.path?')
    else:
        this = sys.modules[__name__]
        for attr in filter(lambda x: not x.startswith('_'), dir(overrides)):
            setattr(this, attr, getattr(overrides, attr))



from helga import settings
from helga.plugins import match


@match(r'cr(\d+)')
def reviewboard(client, channel, nick, message, matches):
    """
    Automatically responds to reviewboard urls if a user mentions a pattern
    like cr####. Requires REVIEWBOARD_URL to exist in settings with formattable
    substring '{review}'
    """
    reviews = [settings.REVIEWBOARD_URL.format(review=cr) for cr in matches]
    return '{0} might be talking about codereview: {1}'.format(nick, ', '.join(reviews))



"""
Utils for deailing with encode/decode and unicode pain
"""
from functools import wraps


def to_unicode(bytestr, errors='ignore'):
    """
    Enforces a byte string to be unicode. If it already is unicode, no
    action is taken. Otherwise, the byte string is decoded as UTF-8
    """
    if isinstance(bytestr, unicode):
        return bytestr
    return bytestr.decode('utf-8', errors)


def from_unicode(unistr, errors='ignore'):
    """
    Enforces a unicode string to be a byte string. If it already is a byte
    string, no action is taken. Otherwise, it is encoded as UTF-8
    """
    if not isinstance(unistr, unicode):
        return unistr
    return unistr.encode('utf-8', errors)


def to_unicode_args(fn):
    """
    Automatically converts all positional byte string arguments to unicode
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        args = list(args)
        for i, val in enumerate(args):
            if isinstance(val, str):
                args[i] = to_unicode(val)
        return fn(*args, **kwargs)
    return wrapper


def from_unicode_args(fn):
    """
    Automatically converts all positional unicode arguments to byte strings
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        args = list(args)
        for i, val in enumerate(args):
            if isinstance(val, unicode):
                args[i] = from_unicode(val)
        return fn(*args, **kwargs)
    return wrapper
